!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 826 $
!> $Author: dsu $
!> $Date: 2022-03-24 10:10:16 -0700 (Thu, 24 Mar 2022) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/cvolume.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!c ----------------------------------------------------------------------
!c subroutine cvolume
!c ------------------
!c
!c compute volumes on a nodal basis for a block discretization
!c
!c written by:      Uli Mayer - June 10, 96 
!c
!c last modified:   Tom Henderson - April 7,2005
!c
!c                  Danyang Su - Sept. 10, 2018
!c                  Unstructured grid and HPC capabilities
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c 
!c                                                                    I O
!c passed:   -
!c
!c common:   
!c gen.f:    real*8:
!c           -------
!c           delx(nvx)          = spatial increment in x-direction    + -
!c           dely(nvy)          = spatial increment in y-direction    + -
!c           delz(nvz)          = spatial increment in z-direction    + -
!c           cvol(nn)           = nodal volumes                       * +
!c
!c           integer*4:
!c           ----------
!c           nvx                = number of control volumes in        + - 
!c                                x-direction
!c           nvy                = number of control volumes in        + -
!c                                y-direction
!c           nvz                = number of control volumes in        + -
!c                                z-direction
!c
!c local:    integer*4:
!c           ----------
!c           ivx                = counter (number of control volumes
!c                                in x-direction)
!c           ivy                = counter (number of control volumes
!c                                in y-direction)
!c           ivz                = counter (number of control volumes
!c                                in z-direction)
!c           ivol               = pointer (current control volume)
!c
!c external: -
!c ----------------------------------------------------------------------
 
      subroutine cvolume
 
#ifdef PETSC
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
      use petscsys
#endif
#endif
      use parm
      use gen

#ifdef PETSC
      use petsc_mpi_common, only : petsc_mpi_finalize
#endif

      implicit none
      
#ifdef PETSC
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif
#endif

#ifdef PETSC
      real*8 :: cvol_tot_gbl
      PetscErrorCode :: ierrcode
#endif

      real*8, parameter  ::               &
          pi = 3.141592653589793d0,       &
          r0 = 0.0d0, r2=2.0d0,           &
          rhalf=0.5d0
      
      integer :: info_dbg, ivol, ivx, ivy, ivz 

      real*8 :: rvol

     
      info_dbg = 0

!c  optional radial coordinates
!c  see infcvs.f
    
      if (discretization_type == 0) then

!c  loop over control volumes  
 
        ivol = 0

        do ivz = nvzgls,nvzgle                  ! number of control volumes in z
          do ivy = nvygls,nvygle                ! number of control volumes in y
            do ivx = nvxgls,nvxgle              ! number of control volumes in x

!c  pointer to current control volume
              ivol = ivol+1

!c  calculate volume
 
              cvol(ivol) = delx(ivx)*dely(ivy)*delz(ivz)

!c correct for radial coord
              if (radial_coord) then
                if (half_cells) then
                  if (ivx .eq. 1) then

!----------------------------------old------------------------------------------
!                  !cvol(ivol) = pi*delx(ivx)*delx(ivx)*delz(ivz)
!                   cvol(ivol) = pi*rhalf*delx(ivx)*delx(ivx)  &
!                                 *delz(ivz)
!                elseif (ivx .eq. nvx) then
!   !                 cvol(ivol) = r2*pi*
!   !  &                  (xg(ivol)-rhalf*delx(ivx))*delx(ivx)*delz(ivz)
!                     cvol(ivol) = r2*pi*xg(ivol)*delx(ivx)*delz(ivz)
!                else ! full cell in xx direction
!                 ! cvol(ivol) = r2*pi*xg(ivol)*delx(ivx)*delz(ivz)
!                   cvol(ivol) = r2*pi*xg(ivol)*rhalf*(delx(ivx-1)+delx(ivx))&
!                                 *delz(ivz)
!----------------------------------old-------------------------------------------                 

!----------------------------------M. Xie, 2012-12-07-----------------------------
                  cvol(ivol) = pi*delx(ivx)*delx(ivx)*delz(ivz)
                  ! cvol(ivol) = pi*rhalf*delx(ivx)*delx(ivx)
  !   &                          *delz(ivz)
                  elseif (ivx .eq. nvxgbl) then
                    cvol(ivol) = r2*pi*                                &
     &                  (xg(ivol)-rhalf*delx(ivx))*delx(ivx)*delz(ivz)
   !                  cvol(ivol) = r2*pi*xg(ivol)*delx(ivx)*delz(ivz)
                  else ! full cell in xx direction
                  cvol(ivol) = r2*pi*xg(ivol)*delx(ivx)*delz(ivz)
                 !  cvol(ivol) = r2*pi*xg(ivol)*rhalf*(delx(ivx-1)+delx(ivx))
  !   &                          *delz(ivz)                 
 !----------------------------------M. Xie, 2012-12-07-----------------------------
                  end if
                else ! full cells
                  cvol(ivol) = r2*pi*xg(ivol)*delx(ivx)*delz(ivz)
                end if
              end if
            continue

            end do                      ! number of increments in x
          end do                        ! number of increments in y
        end do                          ! number of increments in z

      else

      end if

!c calculate total control volume for the whole domain
      cvol_tot = r0
      do ivol = 1,nngl
#ifdef PETSC
        if(node_idx_lg2l(ivol) < 0) then
          cycle
        end if
#endif
        cvol_tot = cvol_tot + cvol(ivol)
      end do

#ifdef PETSC
      call MPI_Allreduce(cvol_tot, cvol_tot_gbl,1,MPI_REAL8,MPI_SUM, &
                         Petsc_Comm_World,ierrcode)
      CHKERRQ(ierrcode)
      cvol_tot = cvol_tot_gbl
#endif

!c  output total control volume to screen and log file
      if (rank == 0 .and. b_enable_output) then
        write(*,'(a,1x,e16.8)') 'Total volume of simulation domain',cvol_tot
        write(*,'(a,1x,e16.8)') 'Total volume of simulation domain',cvol_tot
      end if

      info_dbg = 0
      if (info_dbg .gt. 0) then
        if (b_enable_output .and. b_enable_output_gen) then
          rvol = 0.0d0
          if (discretization_type == 0) then
            write(*,'(2a)')                                         &
                  "  ivol   ivx   ivy   ivz              xg",          &
                  "              yg              zg            cvol"
            ivol = 0
            do ivz = nvzgls,nvzgle                  ! number of control volumes in z
              do ivy = nvygls,nvygle                ! number of control volumes in y
                do ivx = nvxgls,nvxgle              ! number of control volumes in x
                  ivol = ivol+1
                  !write(*,'(4i6,4(1x,es15.7))') ivol,ivx,ivy,ivz,   &
                  !      xg(ivol),yg(ivol),zg(ivol),cvol(ivol)
                  write(*,'(i6,4(1x,es15.7))') ivol,                &
                        xg(ivol),yg(ivol),zg(ivol),cvol(ivol)
                  rvol = rvol + cvol(ivol)
                end do
              end do
            end do
            write(*,'(a,1x,es24.16)') "total volume:",rvol
          else
          end if
        end if
        
        if (info_dbg .gt. 1) then
          if (rank == 0) then
            write(*,*) 'program stopped by dbg in cvolume.f'
          end if
#ifdef PETSC
          call petsc_mpi_finalize
#endif
          stop
        end if
      end if
      info_dbg = 0

      return
      end
