
Program driver_pc

#ifdef PETSC
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
      use petscsys
#endif
#endif

      use parm
      use gen
#ifdef PETSC
      use petsc_mpi_common, only : petsc_mpi_initialize,               &
                                   petsc_mpi_finalize
#endif


      implicit none
#ifdef PETSC
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif
#endif

      integer :: i, ivol, igb, ierrcode, tid, itz
      logical :: bflag
      
      external welcome, opngfls, initcpgs, constnts, sptldisc

!c  initialize MPI parallel environment  
#ifdef PETSC
      call petsc_mpi_initialize(rank, nprcs)
      
      !If run with threadcomm, use the number of threads from '-threadcomm_nthreads'

      call PetscOptionsGetInt(                                         &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                          PETSC_NULL_OPTIONS,PETSC_NULL_CHARACTER,     &
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR == 7)
                          PETSC_NULL_OBJECT,PETSC_NULL_CHARACTER,      &
#else
                          PETSC_NULL_CHARACTER,                        &
#endif
                          '-threadcomm_nthreads',nthreads_per_proc,    &
                          bflag,ierrcode)
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
      CHKERRA(ierrcode)
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 8)
      CHKERRQ(ierrcode)
#endif

      
      !If run without threadcomm, use the number of threads from '-mpicomm_nthreads'
      if (.not. bflag) then

        call PetscOptionsGetInt(                                       &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                            PETSC_NULL_OPTIONS,PETSC_NULL_CHARACTER,   &
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR == 7)
                            PETSC_NULL_OBJECT,PETSC_NULL_CHARACTER,    &
#else
                            PETSC_NULL_CHARACTER,                      &
#endif
                            '-mpicomm_nthreads',nthreads_per_proc,     &
                            bflag,ierrcode)
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
        CHKERRA(ierrcode)
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 8)
        CHKERRQ(ierrcode)
#endif

      end if
      
      if (.not. bflag) then
        nthreads_per_proc = 1 
      else
        if (nthreads_per_proc <1) then
          if (rank == 0) then
            write(*,*) "Error: number of threads < 1"
          end if
          call petsc_mpi_finalize  
          stop
        end if
      end if

      call PetscMemorySetGetMaximumUsage(ierrcode)
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
      CHKERRA(ierrcode)
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 8)
      CHKERRQ(ierrcode)
#endif
#else
      rank = 0
      nprcs = 1
      nthreads_per_proc = 1
      Petsc_Comm_World = 0
      Petsc_Comm_Self = 0
#endif

#ifdef PETSC
      if(nprcs > 1) then
          write(str_rank,'(A,I0)')"_rank_", rank 
      else
          str_rank = ""
      end if
#else
      str_rank = ""
#endif


!c  hardwired logical variables

      show_module = .true.
!!cprovi-------------------------------------------------
!!cprovi We can swith the option of analytic and numerical
!!cprovi derivatives 
!!cprovi-------------------------------------------------
      analyt_deriv_rt = .false.
!!cprovi-------------------------------------------------

      prefix = 'stedvs'
      l_prfx=len_trim(prefix)

!c  open problem specific input file, generic output file
!c  and scratch file for temporary data storage
      call opngfls
      
!c  initialize global control parameters
      call initcpgs
      
      
!c  read spatial discretization parameters, discretize solution domain
!c  and compute nodal volumes
      if (varsat_flow.or.reactive_transport) then
        call sptldisc
      end if

      
!c    finalize petsc solver
#ifdef PETSC
      call petsc_mpi_finalize
#endif

      !c stop
      
End Program
