!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 826 $
!> $Author: dsu $
!> $Date: 2022-03-24 10:10:16 -0700 (Thu, 24 Mar 2022) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/file_unit.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!> module of file unit allocation
!c ----------------------------------------------------------------------
!c
!c automatic get a file unit to open the file
!c
!c written by:      Danyang Su - October 28, 2014
!c
!c last modified:
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c                                                                    I O
!c passed:   integer*4:
!c           ----------
!c           itmp
!c
!c           logical:
!c           ---------
!c           found_name
!c
!c           character:
!c           ----------
!c           name
!c
!c common:   -
!c
!c local:    character:
!c           ----------
!c           string
!c
!c external: -
!c ----------------------------------------------------------------------
module file_unit

#ifdef PETSC
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
      use petscsys
#endif
#endif

    use gen, only : ilog, idbg, rank
    
#ifdef PETSC
    use petsc_mpi_common, only : petsc_mpi_finalize
#endif

    implicit none

#ifdef PETSC
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif
#endif

#ifdef PETSC
      PetscErrorCode :: ierrcode
#endif
    
    public :: lun_get, lun_set, lun_free, lun_available
    integer, parameter :: lun_success = 0             !lun: Logical Unit Number
    integer, parameter :: lun_none_free = -1
    integer, parameter :: lun_not_open = -2
    
    private
    
    integer, parameter :: min_lun = 30000             !File number 1-30000 is reserved for binary output
                                                      !as binary output automatically generate file unit
    integer, parameter :: max_lun = 65545             !Maximum file number of 65536 including input files
    
    logical, dimension(min_lun:max_lun) :: lun_open = .false.
    
    contains
    
    !>
    !> Get file unit number
    !>
    integer function lun_get()
      
      implicit none
      
      integer :: i
      logical :: flag

      flag = .false.
    
      do i = min_lun, max_lun
        if (.not. lun_open(i)) then
          lun_get = i
          lun_open(i) = .true.
          flag = .true.
          exit
        end if
      end do

      if (flag) then
        return
      end if
      
      lun_get = lun_none_free
      
      if (rank == 0) then
        write(*,*) "Error: Get file unit number exceed maximum value"
        write(*,'(a)') "Error: Get file unit number exceed maximum value"
        write(*,'(a)') "       Increase the maximum value and recompile"
        close(ilog)
      end if
      
#ifdef PETSC
      call petsc_mpi_finalize
#endif
      stop
    
    end function lun_get
    
    !>
    !> Set file unit number
    !>
    subroutine lun_set(lun_to_set)
      
      implicit none
      
      integer :: lun_to_set
      
      integer :: i
      
      i = lun_to_set
      
      if (i>=min_lun .and. i<=max_lun) then
        lun_open(i) = .true.
        return
      end if

      if (rank == 0) then
        write(*,*) "Error: Set file unit number exceed maximum value"
        write(*,'(a)') "Error: Set file unit number exceed maximum value"
        write(*,'(a)') "       Increase the maximum value and recompile"
        write(*,'(a,i10)') "Error file unit value ", lun_to_set
        close(ilog)
      end if
      
#ifdef PETSC
      call petsc_mpi_finalize
#endif
      stop
    
    end subroutine lun_set
    
    !>
    !> Free file unit number
    !>
    subroutine lun_free(lun_to_free)
    
      implicit none
      
      integer :: lun_to_free
      integer :: i
      
      i = lun_to_free
      
      if (i>=min_lun .and. i<=max_lun) then      
        if (lun_open(i)) then
          lun_open(i) = .false.
        end if
      end if

      return

      if (rank == 0) then
        write(*,*) "Warning: Free file unit number exceed maximum value"
        write(*,'(a)') "Warning: Free file unit number exceed maximum value"
        write(*,'(a)') "         Increase the maximum value and recompile"
        write(*,'(a,i10)') "Warning file unit value ", lun_to_free
      end if
      
#ifdef PETSC
      call petsc_mpi_finalize
#endif
      stop
    
    end subroutine lun_free
    
    !>
    !> Check if file unit number is occupied
    !> return false if occupied
    !>
    logical function lun_available(lun_to_check)
    
      implicit none
      
      integer :: lun_to_check
      
      integer :: i
      
      i = lun_to_check
      
      if (i>=min_lun .and. i<=max_lun) then      
        if (lun_open(i)) then
          lun_available = .false.
        else
          lun_available = .true.   
        end if
      else
        lun_available = .true.  
      end if

      return

    end function lun_available

end module file_unit
