!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 856 $
!> $Author: dsu $
!> $Date: 2023-02-28 16:33:18 -0800 (Tue, 28 Feb 2023) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/file_utility.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!> module of file utility
!c ----------------------------------------------------------------------
!c
!c
!c written by:      Danyang Su - October 28, 2014
!c
!c last modified:
!c                  Danyang Su - Jan 30, 2019
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c                                                                    I O
!c passed:   integer*4:
!c           ----------
!c           itmp
!c
!c           logical:
!c           ---------
!c           found_name
!c
!c           character:
!c           ----------
!c           name
!c
!c common:   -
!c
!c local:    character:
!c           ----------
!c           string
!c
!c external: -
!c ----------------------------------------------------------------------
module file_utility

    use gen
    use file_unit, only : lun_get, lun_free

    use geometry_definition

    implicit none
    
    !>
    !> interfaces of checking if file is valid to append data directly
    !>
    interface check_rewind_status
      module procedure check_rewind_status_file
      module procedure check_rewind_status_unit
    end interface check_rewind_status

    character(1), parameter :: strcomment = "!"

    contains

    !> Convert upper case to lower case
    subroutine makelowercase(string)

        implicit none

        character(len=*), intent(inout) :: string

        integer :: i,j

        do i = 1,len_trim(string)

            j = ichar(string(i:i))

            if (j >= 65 .AND. j <= 90) then

                j = j + 32
                string(i:i) = char(j)

            end if

        end do

    end subroutine makelowercase
    
    !> number of space separator in a string
    !> merge duplicated one
    function num_char_in_string(strstring, strchar) result (nchar)

      implicit none
      character(len=*), intent(in) :: strstring, strchar
      integer :: nchar

      !c local variable
      character(256) :: strbuffer
      integer :: i
      logical :: bflag
      character(1), parameter :: charspace = ' '

      nchar = 0
      strbuffer = trim(adjustl(strstring))
      call replacecharacter(strbuffer, strchar, charspace)

      bflag = .true.

      do while(bflag)
        i = index(trim(strbuffer),charspace)
        if (i >= 1) then
          nchar = nchar +1
          strbuffer = adjustl(strbuffer(i+1:))
        else
          bflag = .false.
        end if
      end do

    end function num_char_in_string

    !> locate the position of of command
    function findnextline(iunit, strbuffer, irewind, bwithquote_opt) result(bflag)

      implicit none

      integer, intent(in) :: iunit
      character(len=*), intent(in) :: strbuffer
      integer, intent(in) :: irewind
      logical, intent(in), optional :: bwithquote_opt
      logical :: bflag

      !c local variables
      character(2048) :: strbuffernext
      logical :: bwithquote

      bflag = .false.

      if (present(bwithquote_opt)) then
        bwithquote = bwithquote_opt
      else
        bwithquote = .false.
      end if

      !c rewind type, 0: auto, 1: rewind, 2 or others: norewind
      if (irewind == 1) then
        rewind(iunit)
      end if

      do while (readnextline(iunit, strbuffernext, bwithquote))
        if (trim(strbuffernext) == trim(strbuffer)) then
          bflag = .true.
          exit
        end if
      end do

      if (.not. bflag) then
        if (irewind == 0) then
          rewind(iunit)
          do while (readnextline(iunit, strbuffernext, bwithquote))
            if (trim(strbuffernext) == trim(strbuffer)) then
              bflag = .true.
              exit
            end if
          end do
        end if
      end if

    end function findnextline

    !> Read next line, skip comment line start with comment character
    function readnextline(iunit, strbuffer, withquote, lowercase,      &
                          withcomment, original) result (bflag)
    
        implicit none
        
        integer, intent(in) :: iunit
        character(len=*), intent(inout) :: strbuffer
        logical, intent(in), optional :: withquote, lowercase,         &
                                         withcomment, original
        logical :: bflag
        
        !c local variables
        integer :: i, istat, ilen
        logical :: bwithquote, blowercase, bwithcomment, boriginal
        character*1024 :: strbuffer_bk

        if (present(withquote)) then
          bwithquote = withquote
        else
          bwithquote = .false.
        end if
        
        if (present(lowercase)) then
          blowercase = lowercase
        else
          blowercase = .true.
        end if

        if (present(withcomment)) then
          bwithcomment = withcomment
        else
          bwithcomment = .false.
        end if

        if (present(original)) then
          boriginal = original
        else
          boriginal = .false.
        end if

        strbuffer = ""
        bflag = .true.

        do while(bflag)

          read(iunit, "(a)", iostat = istat) strbuffer
          strbuffer_bk = strbuffer
      
          if (istat > 0) then             !Error in reading  
            if (rank == 0 .and. b_enable_output) then
              write(*, *) "Error in reading file."
            end if
            bflag = .false. 
            exit            
          else if (istat < 0) then    !End of file            
            bflag = .false.  
            exit                        
          end if

          strbuffer = adjustl(strbuffer)

          if(len_trim(strbuffer) /= 0 ) then
            
            ! Conver all case to lower case
            if(blowercase) then
              call makelowercase(strbuffer)
            end if
                
            if (.not.bwithcomment) then
              if (strbuffer(1:1) /= strcomment) then
                call replacecharacter(strbuffer, achar(9), " ")     !replace tab with blank space
                if (bwithquote) then
                  !check double quote
                  i = index(strbuffer,'"')
                  if(i == 1) then
                    strbuffer = adjustl(strbuffer(i+1:))
                    i = index(strbuffer,'"')
                    if(i > 0) then
                      strbuffer = strbuffer(:i-1)
                    end if
                  end if
                  
                  !check single quote
                  i = index(strbuffer,"'")
                  if(i == 1) then
                    strbuffer = adjustl(strbuffer(i+1:))
                    i = index(strbuffer,"'")
                    if(i > 0) then
                      strbuffer = strbuffer(:i-1)
                    end if
                  end if
                end if
                exit
              end if
            else
              exit
            end if

          end if

        end do

#ifdef PGI
        ilen = len_trim(strbuffer)
        if (ilen > 0) then
          !c remove 'Carriage Return' in string when PGI compiler is used.
          if (iachar(strbuffer(ilen:ilen)) == 13) then
            strbuffer = strbuffer(:ilen-1)
          end if
        end if
#endif

        if (boriginal) then
          strbuffer = strbuffer_bk
          if (blowercase) then
            call makelowercase(strbuffer)
          end if
        end if
    
    end function readnextline
    
    !> replace character in string
    subroutine replacecharacter(str, stra, strb)
    
        implicit none
        
        character(len=*), intent(inout) :: str
        character(len=*), intent(in) :: stra, strb
        character(2048) :: tempstr 
        
        integer :: i, j, k, n1, n2, n3
        
        n1 = len(str)
        n2 = len(stra)
        n3 = len(strb)        
        j = 1
        k = 0
        do i = 1, n1 - n2 + 1
            if (k>0) then
                k = k - 1
                cycle
            end if
            if(str(i:i+n2-1) == stra(1:n2)) then
                tempstr(j: j + n3 - 1) = strb(1:n3)
                j = j + n3
                k = n2 - 1
            else                
                tempstr(j:j) = str(i:i)
                j = j + 1
            end if
        end do
        
        str = tempstr
        
    end subroutine replacecharacter

    !> positions the specified file to just after the specified restart time
    !> all the transient data should already included the header file
    subroutine find_append_time(iunit,iskip,time_append)

      implicit none

      integer, intent(in) :: iunit
      integer, intent(in) :: iskip
      real*8, intent(inout) :: time_append

      !c local variables
      integer :: i, irecord
      real*8 :: rdummy
      character(len=256) :: strdummy
      logical :: bflag1, bflag2, bflag3
      real*8, parameter :: r0 = 0.0d0, rsmall = 1.0d-4

      time_append = r0

      rewind(iunit,err=10)

      irecord = 0
      rdummy = r0

      bflag1 = .false.
      bflag2 = .false.
      bflag3 = .false.

      !skip version line
      if (b_writeversion_tecplot) then
        read(iunit,*,err=10,end=10) strdummy
      end if

      !skip the three header line
      do while(.true.)
        !c read format * only read the string before the first space,
        !c do not use '(a)' here which reads the whole line
        read(iunit,*,err=10,end=10) strdummy
        if (trim(strdummy) == "title") then
          bflag1 = .true.
        else if (trim(strdummy) == "variables") then
          bflag2 = .true.
        else if (trim(strdummy) == "zone") then
          bflag3 = .true.
        end if
        if (bflag1 .and. bflag2 .and. bflag3) then
          exit
        end if
      end do

      if (time_io_rs > r0) then
        do while(.true.)
          read(iunit,*,err=10,end=10) rdummy
          if (rdummy >= time_io_rs .or. abs(rdummy-time_io_rs) < rsmall) then
            irecord = 1
            backspace(iunit,err=10)
            exit
          end if
        end do
        if (irecord == 1) then
          do i = 1, iskip-1
            read(iunit,*,err=10,end=10) rdummy
          end do
          time_append=rdummy
        end if

      end if

10    return

    end subroutine find_append_time

    !>
    !> reposition to the first record after header (comment lines)
    !>
    subroutine rewind_first_record(iunit)

      implicit none

      integer, intent(in) :: iunit

      character(2048) :: strbuffer2048

      if(readnextline(iunit, strbuffer2048, withquote=.false.)) then
        backspace(iunit)
      end if

    end subroutine rewind_first_record

    !> positions the specified file to just after the specified restart time
    !> all the transient data should already included the header file
    !> numerical error is possible if the output resolution is not sufficient for time
    !> due to round error, for example, 1.00123e3 and 1.00125e3 may be written out as
    !> 1.0012e3 and 1.0012e3.
    subroutine reposition_file(iunit,irecord,time_opt)

      implicit none

      integer, intent(in) :: iunit
      integer, intent(inout) :: irecord
      real*8, optional :: time_opt

      !c local variables
      integer :: i
      real*8 :: rdummy, rdummy_old, time_lot
      character(len=256) :: strdummy
      logical :: bflag1, bflag2, bflag3, bflag4
      real*8, parameter :: r0 = 0.0d0, rsmall = 1.0d-4

      rewind(iunit,err=20)

      if (present(time_opt)) then
        time_lot = time_opt
      else
        time_lot = time_io
      end if

      irecord = 0
      rdummy = r0

      bflag1 = .false.
      bflag2 = .false.
      bflag3 = .false.

      !skip version line
      if (b_writeversion_tecplot) then
        read(iunit,*,err=10,end=10) strdummy
      end if

      !skip the three header line
      do while(.true.)
        !c read format * only read the string before the first space,
        !c do not use '(a)' here which reads the whole line
        read(iunit,*,err=10,end=10) strdummy
        if (trim(strdummy) == "title") then
          bflag1 = .true.
        else if (trim(strdummy) == "variables") then
          bflag2 = .true.
        else if (trim(strdummy) == "zone") then
          bflag3 = .true.
        end if
        if (bflag1 .and. bflag2 .and. bflag3) then
          exit
        end if
      end do

      !c flag to check if end of file is reached
      bflag4 = .false.

      if (time_lot > r0) then
        bflag4 = .true.
        do while(.true.)
          rdummy_old = rdummy
          read(iunit,*,err=10,end=10) rdummy
          irecord = irecord + 1
          if (rdummy >= time_lot) then
            backspace(iunit,err=10)
            bflag4 = .false.
            exit
          end if
        end do
      else
        return
      end if

10    continue

      !c The last time record is less than restart time, 
      !c the record position is now at EOF thus extra backspace is needed.                                                    
      if (bflag4) then
          backspace(iunit,err=20)
      end if

      !c When simulation is not normal exit, some of the cached data are not written back to the file.
      !c For transient output, the last record may be not at the same time.
      !c The following code is for the file with last record time earlier than restart time.
      if (irecord == 0 .or. rdummy < time_lot) then
        backspace(iunit)
        inquire(unit=iunit,name=strdummy)
        if (rank == 0 .and. b_enable_output) then
          !DSU: restart time is very close to last record, which may be caused by I/O format.
          !DSU: consider absolute difference and relative difference
          if ((irecord <= 1 .and. time_io_rs-rdummy > rsmall) .or. &
              (irecord > 1 .and. (time_io_rs-rdummy)/rdummy > rsmall)) then
            write(ifls,'(/72a)') ('-',i=1,72)
            write(ifls,'(3(a,1x))') "warning - last data record in file",&
                  trim(strdummy),"is before restarted time"
          end if
        end if
      end if

20    return

    end subroutine reposition_file

    !> check if the file exist with tecplot header
    function check_rewind_status_file(strfile) result(bflag)

      implicit none

      character(len=*) :: strfile
      logical :: bflag

      !c local variable
      logical :: is_exist
      integer :: iunit
      character(256) :: strbuffer
      logical :: bflag1, bflag2, bflag3

      bflag = .false.

      bflag1 = .false.
      bflag2 = .false.
      bflag3 = .false.

      inquire(file=trim(strfile),exist=is_exist)

      if (is_exist) then
        iunit = lun_get()
        open(iunit,file=trim(strfile),status='unknown')

        if (b_writeversion_tecplot) then
          read(iunit,'(a1)',err=10,end=10) strbuffer
        end if

        do while(.true.)
          !c read format * only read the string before the first space,
          !c do not use '(a)' here which reads the whole line
          read(iunit,*,err=10,end=10) strbuffer
          if (trim(strbuffer) == "title") then
            bflag1 = .true.
          else if (trim(strbuffer) == "variables") then
            bflag2 = .true.
          else if (trim(strbuffer) == "zone") then
            bflag3 = .true.
          end if
          if (bflag1 .and. bflag2 .and. bflag3) then
            bflag = .true.
            exit
          end if
        end do

10    continue
        close(iunit)
        call lun_free(iunit)

      end if

    end function check_rewind_status_file

    !> check if the file exist with tecplot header
    function check_rewind_status_unit(iunit) result(bflag)

      implicit none

      integer, intent(in) :: iunit
      logical :: bflag

      !c local variable
      logical :: is_exist, is_opened
      logical :: bflag1, bflag2, bflag3
      character(256) :: strbuffer

      bflag = .false.

      is_exist = .false.
      is_opened = .false.

      bflag1 = .false.
      bflag2 = .false.
      bflag3 = .false.

      inquire(unit=iunit,exist=is_exist,opened=is_opened)

      if (is_exist .and. is_opened) then
        if (b_writeversion_tecplot) then
          read(iunit,'(a1)',err=10,end=10) strbuffer
        end if

        do while(.true.)
          !c read format * only reads the string before the first space,
          !c do not use '(a)' here which reads the whole line
          read(iunit,*,err=10,end=10) strbuffer
          if (trim(strbuffer) == "title") then
            bflag1 = .true.
          else if (trim(strbuffer) == "variables") then
            bflag2 = .true.
          else if (trim(strbuffer) == "zone") then
            bflag3 = .true.
          end if
          if (bflag1 .and. bflag2 .and. bflag3) then
            bflag = .true.
            exit
          end if
        end do

      end if

10    continue

      if (is_exist .and. is_opened) then
        rewind(iunit)
      end if

    end function check_rewind_status_unit

end module
