!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 850 $
!> $Author: dsu $
!> $Date: 2023-01-27 08:58:23 -0800 (Fri, 27 Jan 2023) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/initcpgs.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!c ----------------------------------------------------------------------
!c subroutine initcpgs
!c -------------------
!c
!c control parameters for global system
!c modified for optional density dependent flow simulation
!c
!c written by:      Uli Mayer - May 9, 96
!c
!c last modified:   Tom Henderson - August 21, 2002
!c
!c            Celine Blitz Frayret (CBF) for Frederic Gerard, December 14, 2018 
!c            Added root_uptake global control parameter      
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c 
!c                                                                    I O
!c passed:   -
!c
!c common:   
!c gen.f:    integer*4:
!c           ----------
!c           idat               = unit number, run specific input     + -
!c                                             file
!c           igen               = unit number, generic output file    + -
!c           ilog               = unit number, log book               + -
!c           itmp               = unit number, temporary storage      + -
!c           l_time_unit        = length of time_unit for output      * +
!c
!c           logical:
!c           --------
!c           fully_saturated    = .true.  -> saturated conditions     * +
!c           geo_chemistry      = .true.  -> local or background and  * +
!c                                           source chemistry
!c           reactive_transport = .true.  -> perform reactive         * +
!c                                           transport simulation
!c           steady_flow        = .true.  -> steady state flow        * +
!c           transient_flow     = .true.  -> .not.steady_flow,        * +
!c                                        -> transient flow
!c           variably_saturated = .true.  -> .not.fully_saturated,    * +
!c                                        -> variably saturated
!c                                           conditions
!c           varsat_flow        = .true.  -> perform flow simulation  * +
!c         root_uptake        = .true.  -> calculate root water
!c                                           uptake  CBF
!c
!c           character:
!c           ----------
!c           problem_title      = problem title                       * +
!c           section_header     = section header                      * +
!c           drive              = drive of program installation       * +
!c           time_unit          = time unit for output -> 'years'     * +
!c                                                        'days'
!c                                                        'hours'
!c                                                'seconds'
!c dens.f:   logical:
!c           --------
!c           density_dependence = .true.  -> density-dependant flow 
!c
!c local:    integer*4:
!c           ----------
!c           l_string           = length of text string
!c
!c           logical:
!c           --------
!c           found_section      = .true.  -> section header was
!c                                           found in input file
!c           found_subsection   = .true.  -> subsection header was
!c                                           found in input file
!c
!c           character:
!c           ----------
!c           subsection         = name of subsection in input file
!c
!c
!c external: readbloc  = read section of input file and write to
!c                       temporary file
!c ----------------------------------------------------------------------
 
      subroutine initcpgs
 
      use parm
      use gen
      use file_utility, only : replacecharacter
#ifdef PETSC
      use petsc_mpi_common, only : petsc_mpi_finalize
#endif

      implicit none
      
      integer :: i, l_string, ierrcd

      external readbloc
 
      real*8 :: mem_input
      logical found_section, found_subsection,iserror
      character*72 subsection, subkeyword

!c     set defaults
      ierrcd = 0
      drive = 'u'

!c  read numerical data and write to temporary file   
   
      section_header = 'global control parameters'
      call readbloc (idat,itmp,section_header,found_section,.true.)

!c  define length of section header

      l_string = index(section_header,'  ')-1
      if (l_string.eq.-1.or.l_string.gt.72) then
         l_string=72
      end if

!c  terminate program if section header not found

      if (.not.found_section) then
        if (rank == 0) then
          write(*,*) 'error reading input file'
          write(*,*) 'section "',section_header(:l_string),'" missing'
          close(ilog)
        end if
#ifdef PETSC
        call petsc_mpi_finalize
#endif
        stop
      end if

!c  write section header to current output file
      if (b_enable_output .and. b_enable_output_gen) then
        write(*,'(/72a)')('-',i=1,72)
        write(*,'(a)') section_header(:l_string)
        write(*,'(72a/)')('-',i=1,72)
      end if

!c  read global control parameters 

      root_uptake = .false.                           !DSU - This is an optional parameter, initialize with false if the value is not provided
      ierrcd = 1
      read(itmp,*,err=999,end=999) problem_title
      read(itmp,*,err=999,end=999) varsat_flow
      read(itmp,*,err=999,end=999) steady_flow
      read(itmp,*,err=999,end=999) fully_saturated
      read(itmp,*,err=999,end=999) reactive_transport
      read(itmp,*,err=998,end=999) root_uptake        !CBF - FG 07-2017 - fast trick to enable reading of trunk benchmarks
998   continue

     
!c  number of time steps to skip for writing log file (and on screen)
      skip_time = 0   !initially none
      subsection = 'skipped time steps in log file'
         
      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        ierrcd = 2
        read(itmp,*,err=999,end=999) skip_time
      
        if (rank == 0 .and. b_enable_output) then
          write(*,*) 'skipped time steps in log file: ', skip_time
          write(*,*) 'skipped time steps in log file: ', skip_time
        end if

      end if                 !(found_section)

!c  type of redox reactions

      subsection = 'specify drive of program installation'

      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        ierrcd = 3
        read(itmp,*,err=999,end=999) drive
      end if
      
!c  write version information to tecplot data file or not

      subsection = 'write version to tecplot data file'

      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
          b_writeversion_tecplot = .true.
      else
          b_writeversion_tecplot = .false.
      end if

!c  convert external data file to binary format.
!c  for large dataset, reading binary format is much faster 
!c  than ascii format line by line.
!c      export_ascii_binary = .false.
!c      subsection = 'export external data to binary format'
!c      call findstrg(subsection,itmp,found_subsection)
!c      if (found_subsection) then
!c        export_ascii_binary = .true.
!c      end if
      
#ifdef PETSC
      subsection = 'use distributed restart files to read'

      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        b_distributed_restart_read = .true.
      else
        b_distributed_restart_read = .false.
      end if
#else
      b_distributed_restart_read = .false.
#endif
      
#ifdef PETSC
      subsection = 'use distributed restart files to write'

      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        b_distributed_restart_write = .true.
      else
        b_distributed_restart_write = .false.
      end if
#else
      b_distributed_restart_write = .false.
#endif
      
      subsection = 'use binary restart file to read'
      call findstrg(subsection,itmp,found_subsection)
#ifdef PETSC
      if (found_subsection) then
        b_binary_restart_read = .true.
      else
#endif
        if (found_subsection) then
          if (rank == 0) then
            write(*,'(a)') 'Binary restart file reading is only available when built with MPI'
          end if
        end if
        b_binary_restart_read = .false.
#ifdef PETSC
      end if
#endif
      
      subsection = 'use binary restart file to write'
      call findstrg(subsection,itmp,found_subsection)
#ifdef PETSC
      if (found_subsection) then
        b_binary_restart_write = .true.
      else
#endif
        if (found_subsection) then
          if (rank == 0) then
            write(*,'(a)') 'Binary restart file writing is only available when built with MPI'
          end if
        end if
        b_binary_restart_write = .false.
#ifdef PETSC
      end if
#endif

!c  parallel reading method for initial condition, material properties and etc.
      subsection = 'read spatial data using master processor'
      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        read_spatial_master_proc = .true.
      else
        read_spatial_master_proc = .false.
      end if

      subsection = 'ignore restart time'

      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        ignore_restart_time = .true.
      else
        ignore_restart_time = .false.
      end if

      subsection = 'backup frequency'

      call findstrg(subsection,itmp,found_subsection)

      if (found_subsection) then
        ierrcd = 5
        read(itmp,*,err=999,end=999) backup_frequency
      else
        backup_frequency = 100
      end if

!c  ascii output data format
      subsection = 'use double precision'  
      call findstrg(subsection,itmp,found_subsection)
      if (found_subsection) then
        ascii_fmt = '(1000(1pe25.17e3))'
        ascii_fmt_iir = '(2(i10,1x),1000(1pe25.17e3))'
      else
        ascii_fmt = '(1000(1pe15.6e3))'
        ascii_fmt_iir = '(2(i10,1x),1000(1pe15.6e3))'
      end if

!c  maximum memory avaiable for master node per CPU
      subsection = 'memory limitation per processor'  
      call findstrg(subsection,itmp,found_subsection)
      if (found_subsection) then
        ierrcd = 6
        read(itmp,*,err=999,end=999) mem_input, mem_unit
        if (mem_unit == 'MB') then
          !c default, do nothing
        else if (mem_unit == 'GB') then
          mem_input = mem_input*1.0d3
        else if (mem_unit == 'TB') then
          mem_input = mem_input*1.0d6
        else if (mem_unit == 'PB') then
          mem_input = mem_input*1.0d9
        else
          if (rank == 0) then
            write(*,*) 'Error in reading memory limitation per processor'
            write(*,*) 'Error in reading memory limitation per processor'
          end if
          ierrcd = 7
          goto 999
        end if
        mem_limit = int(mem_input)
      else
        mem_limit = 2**30    !maximum value of interger*4 is 2**31-1
      end if

 
!c  assign related global control parameters
 
      if ((varsat_flow).and.(reactive_transport)) then
        geo_chemistry = .true.
      elseif ((varsat_flow).and.(.not.reactive_transport)) then
        geo_chemistry = .false.
      elseif ((.not.varsat_flow).and.(reactive_transport)) then
        geo_chemistry = .true.
      elseif ((.not.varsat_flow).and.(.not.reactive_transport)) then
        geo_chemistry = .true.
      end if

      transient_flow = .not.steady_flow
      variably_saturated = .not.fully_saturated

      if (steady_flow.and.(.not.reactive_transport)) then
        time_unit = '-'
        l_time_unit = 1
      end if

!cdsu--------------------------------------------------------  
!cdsu   Parallel solver configuration file
!cdsu--------------------------------------------------------
      if (.not. b_solver_cfg_arg) then
        subsection = 'parallel solver configuration file'
        call findstrg(subsection,itmp,found_subsection)
        if (found_subsection) then
          ierrcd = 9
          read(itmp,*,err=999,end=999) str_solvercfg
          !replace '\' with '/'
          call replacecharacter(str_solvercfg,achar(92),achar(47))
        else
          str_solvercfg = ""
        end if
      end if
        
!cdsu--------------------------------------------------------  
!cdsu   Output or not
!cdsu   If true, all the output will be disabled. 
!cdsu   Only the final output "normal exit" is kept.
!cdsu   This section has been moved to opngfls subroutine
!cdsu--------------------------------------------------------         
        !b_enable_output = .true.
        !subsection = 'disable output'        
        !call findstrg(subsection,itmp,found_subsection)
        !
        !if (found_subsection) then
        !  b_enable_output  = .false. 
        !end if 


!cprovi--------------------------------------------------------
!cprovi--------------------------------------------------------
!cprovi--------------------------------------------------------    

      goto 1000

999   continue
      if (rank == 0) then
        write(*,*) 'error reading input file, error code ', ierrcd
        write(*,*) 'section "',section_header(:l_string),'"'
        close(ilog)
      end if
#ifdef PETSC
      call petsc_mpi_finalize
#endif
      stop

1000  return
      end
