!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 826 $
!> $Author: dsu $
!> $Date: 2022-03-24 10:10:16 -0700 (Thu, 24 Mar 2022) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/opngfls.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!c ----------------------------------------------------------------------
!c subroutine opngfls
!c ------------------
!c
!c open general problem specific files 
!c - problem specific input file
!c - generic output file
!c - debugging file
!c - scratch file
!c
!c written by:      Uli Mayer - May 7, 96
!c
!c last modified:   Uli Mayer - November 14, 96
!c                  Tom Henderson - July 21,2006
!c                  Open scratch velocity file
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c 
!c                                                                    I O
!c passed:   -
!c
!c common:   
!c gen.f:    integer*4:
!c           ----------
!c           idat               = unit number, run specific input     * +
!c                                             file
!c           idbg               = unit number, debugging file         * +
!c           idelt              = unit mumber, magnitude of current   * +
!c                                             time step
!c           ifls               = unit number, file information       * +
!c           igen               = unit number, generic output file    * +
!c           ilog               = unit number, log file               * +
!c           icnv               = unit number, data conversion        * +
!c           ihyc               = unit number, initial hydraulic      * +
!c                                             conductivity 
!c                                             distribution
!c           itmp               = unit number, temporary storage      * +
!c           l_prfx             = length of prefix of I/O files       + -
!c
!c           logical:
!c           --------
!c           log_file           = .true.  -> write log file           * +
!c
!c           character:
!c           ----------
!c           prefix             = prefix name for all I/O files       + -
!c
!c local:    -
!c
!c external: -
!c ----------------------------------------------------------------------
 
      subroutine opngfls
 
      use parm
      use gen
      use file_unit, only : lun_get
 
      implicit none
      
      integer :: i
      character*72 :: strblockname
      character*72 :: subsection
      logical :: found_subsection
      
      external findstrginblock

!c  write log file or write log to screen

      log_file = .true.

!c  specify unit numbers
 
      !idat  = 10
      !itmp  = 11 
      !icnv  = 12
      !igen  = 20 
      !ifls  = 21
      !idbg  = 22
      idat  = lun_get()
      itmp  = lun_get() 
      icnv  = lun_get()
      igen  = lun_get() 
      ifls  = lun_get()
      idbg  = lun_get()
      
      !DSU move ifls to initcpgs to enable restart output information

      if (log_file) then
        !ilog  = 23             !log file
        ilog = lun_get()
      else
        ilog  = 6              !screen
      end if

      !ihyc = 25
      !ipor = 26
      !itor = 28
      !istor = 29
      !iskempton = 50 
      !ienergy = 51 
      
!c  open files

!c  problem specific input file 
      !c to be checked later
      write(*,*) 'input file: ', prefix(:l_prfx)//'.dat'

      open(idat,file=prefix(:l_prfx)//'.dat',status='old')
      
!c  check if output is disabled in the input file
      strblockname = 'global control parameters' 
      
      b_enable_output = .true.
      subsection = 'disable output'        
      call findstrginblock(strblockname,subsection,idat,found_subsection)
      
      if (found_subsection) then
        b_enable_output  = .false. 
      end if 
      
      b_enable_output_gen = .true.
      subsection = 'disable output of all gen files'        
      call findstrginblock(strblockname,subsection,idat,found_subsection)      
      if (found_subsection) then
        b_enable_output_gen  = .false. 
      end if 
      
      subsection = 'enable output of master gen file'      
      if (rank == 0) then
        call findstrginblock(strblockname,subsection,idat,found_subsection)      
        if (found_subsection) then
          b_enable_output_gen  = .true. 
        end if 
      end if

#ifdef MEMORY_MONITOR
      b_output_mem_detail = .false.
      subsection = 'enable output of detailed memory usage'
      call findstrginblock(strblockname,subsection,idat,found_subsection)
      if (found_subsection) then
        b_output_mem_detail = .true.
      end if
      if (b_output_mem_detail) then
        ifile_mem = lun_get()
      end if
#endif
 
!c  scratch file for temporary storage 
 
      open(itmp,file=prefix(:l_prfx)//'_o'//trim(adjustl(str_rank))//  &
                '.tmp',status='unknown', form='formatted')

!c  scratch file for data conversion and temporary storage

!cff    open(icnv,status='scratch',form='formatted')
      open(icnv,file=prefix(:l_prfx)//'_o'//trim(adjustl(str_rank))//  &
                '.cnv', status='unknown', form='formatted')
 

!c  debugging file
#ifdef DEBUG
      if(b_enable_output) then
        open(idbg,file=prefix(:l_prfx)//'_o'//trim(adjustl(str_rank))//&
                  '.dbg',status='unknown', form='formatted')
      end if
#endif


      return
      end
