!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 850 $
!> $Author: dsu $
!> $Date: 2023-01-27 08:58:23 -0800 (Fri, 27 Jan 2023) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/readbloc.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!c ----------------------------------------------------------------------
!c subroutine readbloc
!c -------------------
!c
!c this subroutine reads a block from the input file unit nin 
!c the block has to start with section_header='string' and has to end 
!c with 'done' data is written to unit nout for further processing in 
!c the main program 
!c comment lines starting with ! are skipped  
!c both units are rewind after the procedure is finished
!c
!c written by:      Uli Mayer - July, 95
!c
!c last modified:
!c
!c                  Danyang Su - Sept. 10, 2018
!c                  Add support for long input string *256, e.g., database path
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c                                                                    I O
!c passed:   integer*4:
!c           ---------- 
!c           nin                = id number of input file             + -
!c           nout               = id number of output file            + -
!c
!c           character:
!c           ----------
!c           section_header     = character string defining block     + -
!c                                in input file
!c
!c           found              = .true.  -> section header was       * +
!c                                           found in input file
!c           isrewind          
!c
!c common:   -
!c
!c local:    character:
!c           ----------
!c           dummy1             = dummy character string
!c           dummy2             = dummy character string
!c
!c external: -
!c ----------------------------------------------------------------------
      
      subroutine readbloc (nin,nout,section_header,found,     &
     &                     isrewind)
      
      use gen, only : rank, ilog
#ifdef PETSC
      use petsc_mpi_common, only : petsc_mpi_finalize
#endif
      implicit none
      
      logical, intent(out)          :: found
      integer, intent(in)           :: nin,nout
      logical, intent(in)           :: isrewind
      character(len=*), intent(in)  :: section_header
      
!c  change from 72 to 256 to support long string, e.g., database full path
      character*2048                 :: dummy1, dummy2


      found = .false.
      
    
 
!c  rewind unit nout from previous read-procedure in main routine
      if (isrewind) then
       rewind nin
       rewind nout
      else 
       rewind nout
      end if 
!c  read and write procedure
 
100   continue
      read(nin,*,end=400,err=999) dummy1
      if (dummy1.eq.section_header) then
        found = .true.
200     continue
        read(nin,*,end=400,err=999) dummy1
        backspace(nin) 
        read(nin,'(a)') dummy2
        if (dummy2(1:1).ne.'!') then
          write(nout,'(a)') trim(dummy2)
        end if
        if (dummy1.eq.'done') then
          goto 300
        end if
        goto 200
      else
        goto 100
      end if
 
300   continue
 

      if (isrewind) then
       rewind nin
       rewind nout
      else 
       rewind nout
      end if 
      
400   return 

999   continue
      backspace(nin)
      read(nin,'(a)') dummy1
      if (rank == 0) then
        write(*,*) 'SIMULATION TERMINATED'
        write(*,'(2a)') 'error reading in sector: ',trim(section_header)
        write(*,'(2a)') 'error reading in input file: ',trim(dummy1)
        write(*,*) 'SIMULATION TERMINATED'
        write(*,'(2a)') 'error reading in sector: ',trim(section_header)
        write(*,'(2a)') 'error reading in input file: ',trim(dummy1)
        close(ilog)
      end if
#ifdef PETSC
      call petsc_mpi_finalize
#endif
      stop

      end
    
