!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 856 $
!> $Author: dsu $
!> $Date: 2023-02-28 16:33:18 -0800 (Tue, 28 Feb 2023) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/solver/solver_ddmethod.F90 $
!---------------------------------------------------------------------
!********************************************************************!

#ifdef PETSC

!> module: solver_dd
!>
!> written by: Danyang Su
!>
!> module description:
!>
!> Module of solver for domain decomposition method  
!>
!> Note:  
!> See http://www.mcs.anl.gov/petsc/ for detail


module solver_dd

#include <petscversion.h>
    use solver_snes_common  
    
    implicit none  

    integer, parameter :: n_binary_limit = 1024          !2147483647

    contains
    
    !>
    !> Create distributed array (DMDA) for flow
    !> problem
    !>
    subroutine solver_dd_DMDACreate_flow_heat(bflag_flow)
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscdmda.h>
      use petscdmda
#endif

      use gen, only: nvxgbl, nvygbl, nvzgbl, idbg, ilog, rank, nprcs,  &
                     mpiarray_ndim, mpiarray_sizes_gbl,                &
                     mpiarray_sizes_sub, mpiarray_starts_sub,          &
                     mpiarray_sizes_vel_gbl, mpiarray_sizes_vel_sub,   &
                     mpiarray_starts_vel_sub, discretization_type

      use petsc_mpi_common, only : petsc_mpi_finalize

      implicit none

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscdmdef.h>
#include <petsc/finclude/petscdm.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscdmda.h90>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscdmdef.h>
#include <finclude/petscdm.h>
#include <finclude/petscdmda.h>
#include <finclude/petscdmda.h90>
#endif

      PetscBool :: bflag_flow

      type(userctx), pointer :: dmda_selec
      Vec, pointer :: x_selec
      Vec, pointer :: b_selec
      Vec, pointer :: r_selec
      Vec, pointer :: x_loc_selec
      Vec, pointer :: b_loc_selec
      Vec, pointer :: r_loc_selec
        
      PetscErrorCode :: ierr

      PetscInt :: i, istart, iend, info_debug

      !c point to flow or decoupled heat variables
      if (bflag_flow) then
        dmda_selec => dmda_flow
        x_selec => x_flow
        b_selec => b_flow
        r_selec => r_flow
        x_loc_selec => x_flow_loc
        b_loc_selec => b_flow_loc
        r_loc_selec => r_flow_loc
      else
        dmda_selec => dmda_heat
        x_selec => x_heat
        b_selec => b_heat
        r_selec => r_heat
        x_loc_selec => x_heat_loc
        b_loc_selec => b_heat_loc
        r_loc_selec => r_heat_loc
      end if

      !> Initialize
      dmda_selec%dof = 1

      dmda_selec%swidth = stencil_width

      !c structured grid
      if (discretization_type == 0) then
        dmda_selec%dim  = 0

        dmda_selec%xs   = 0
        dmda_selec%xe   = 0
        dmda_selec%xl   = 1
        dmda_selec%gxs  = 0
        dmda_selec%gxe  = 0
        dmda_selec%gxl  = 1

        dmda_selec%ys   = 0
        dmda_selec%ye   = 0
        dmda_selec%yl   = 1
        dmda_selec%gys  = 0
        dmda_selec%gye  = 0
        dmda_selec%gyl  = 1

        dmda_selec%zs   = 0
        dmda_selec%ze   = 0
        dmda_selec%zl   = 1
        dmda_selec%gzs  = 0
        dmda_selec%gze  = 0
        dmda_selec%gzl  = 1

        dmda_selec%nvxgbl   = 1
        dmda_selec%nvygbl   = 1
        dmda_selec%nvzgbl   = 1

        dmda_selec%npx  = 1
        dmda_selec%npy  = 1
        dmda_selec%npz  = 1
        
        !> Initialize mpi subarray parameters
        if (bflag_flow) then
          mpiarray_ndim = 1
        
          mpiarray_sizes_gbl = 1
          mpiarray_sizes_sub = 1
          mpiarray_starts_sub = 0
        
          mpiarray_sizes_vel_gbl = 1
          mpiarray_sizes_vel_sub = 1
          mpiarray_starts_vel_sub = 0
        end if
   
        !> Create DMDA and get local grid boundary
        !3D domain
        if (nvxgbl > 1 .and. nvygbl > 1 .and. nvzgbl > 1) then  
!cdsu Note: DMDABoundaryType has become DMBoundaryType, and all the 
!cdsu enumerationvalues have also been renamed from PETSc 3.5
!cdsu For the windows version, we still use PETSc 3.4.

          call DMDACreate3d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                    DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,             &
                    DMDA_BOUNDARY_NONE,                                &
#else
                    DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                 &
                    DM_BOUNDARY_NONE,                                  &
#endif

                    DMDA_STENCIL_BOX,nvxgbl,nvygbl,nvzgbl,             &
                    PETSC_DECIDE,PETSC_DECIDE,                         &
                    PETSC_DECIDE,                                      &
                    dmda_selec%dof, dmda_selec%swidth,                 &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,dmda_selec%da,ierr) 
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                   dmda_selec%nvxgbl,dmda_selec%nvygbl,                &
                   dmda_selec%nvzgbl,                                  &
                   dmda_selec%npx,dmda_selec%npy,                      &
                   dmda_selec%npz, PETSC_NULL_INTEGER,                 &
                   PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,              &
                   PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,              &
                   PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%xs,             &
                   dmda_selec%ys,dmda_selec%zs,                        &
                   dmda_selec%xl,dmda_selec%yl,                        &
                   dmda_selec%zl,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gxs,       &
                   dmda_selec%gys,dmda_selec%gzs,                      &
                   dmda_selec%gxl,dmda_selec%gyl,                      &
                   dmda_selec%gzl,ierr)
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 3            
            mpiarray_sizes_gbl(1) = nvxgbl            
            mpiarray_sizes_sub(1) = dmda_selec%xl
            mpiarray_starts_sub(1) = dmda_selec%xs
            
            mpiarray_sizes_gbl(2) = nvygbl          
            mpiarray_sizes_sub(2) = dmda_selec%yl
            mpiarray_starts_sub(2) = dmda_selec%ys
            
            mpiarray_sizes_gbl(3) = nvzgbl
            mpiarray_sizes_sub(3) = dmda_selec%zl
            mpiarray_starts_sub(3) = dmda_selec%zs
            
            mpiarray_sizes_vel_gbl(1) = nvxgbl-1 
            if(dmda_selec%xs == dmda_selec%gxs) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%xs
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl 
              mpiarray_starts_vel_sub(1) = dmda_selec%xs-1
            end if
            
            mpiarray_sizes_vel_gbl(2) = nvygbl-1
            if(dmda_selec%ys == dmda_selec%gys) then
              mpiarray_sizes_vel_sub(2) = dmda_selec%yl-1  
              mpiarray_starts_vel_sub(2) = dmda_selec%ys
            else
              mpiarray_sizes_vel_sub(2) = dmda_selec%yl 
              mpiarray_starts_vel_sub(2) = dmda_selec%ys-1
            end if
            
            mpiarray_sizes_vel_gbl(3) = nvzgbl-1 
            if(dmda_selec%zs == dmda_selec%gzs) then
              mpiarray_sizes_vel_sub(3) = dmda_selec%zl-1  
              mpiarray_starts_vel_sub(3) = dmda_selec%zs
            else
              mpiarray_sizes_vel_sub(3) = dmda_selec%zl 
              mpiarray_starts_vel_sub(3) = dmda_selec%zs-1
            end if
          end if
            
        !2D xy domain                                                  
        else if (nvxgbl > 1 .and. nvygbl > 1) then    
          call DMDACreate2d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                  DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,               &
#else
                  DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                   &
#endif
                  DMDA_STENCIL_BOX,                                    &
                  nvxgbl,nvygbl,PETSC_DECIDE,PETSC_DECIDE,             &
                  dmda_selec%dof, dmda_selec%swidth,                   &
                  PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,               &
                  dmda_selec%da,ierr) 
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                    dmda_selec%nvxgbl,dmda_selec%nvygbl,               &
                    PETSC_NULL_INTEGER,                                &
                    dmda_selec%npx,dmda_selec%npy,                     &
                    PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,            &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%xs,             &
                   dmda_selec%ys,PETSC_NULL_INTEGER,                   &
                   dmda_selec%xl,dmda_selec%yl,                        &
                   PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gxs,       &
                   dmda_selec%gys,PETSC_NULL_INTEGER,                  &
                   dmda_selec%gxl,dmda_selec%gyl,                      &
                   PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 2            
            mpiarray_sizes_gbl(1) = nvxgbl            
            mpiarray_sizes_sub(1) = dmda_selec%xl
            mpiarray_starts_sub(1) = dmda_selec%xs
            
            mpiarray_sizes_gbl(2) = nvygbl          
            mpiarray_sizes_sub(2) = dmda_selec%yl
            mpiarray_starts_sub(2) = dmda_selec%ys
            
           
            mpiarray_sizes_vel_gbl(1) = nvxgbl-1 
            if(dmda_selec%xs == dmda_selec%gxs) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%xs
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl 
              mpiarray_starts_vel_sub(1) = dmda_selec%xs-1
            end if
            
            mpiarray_sizes_vel_gbl(2) = nvygbl-1
            if(dmda_selec%ys == dmda_selec%gys) then
              mpiarray_sizes_vel_sub(2) = dmda_selec%yl-1  
              mpiarray_starts_vel_sub(2) = dmda_selec%ys
            else
              mpiarray_sizes_vel_sub(2) = dmda_selec%yl 
              mpiarray_starts_vel_sub(2) = dmda_selec%ys-1
            end if
          end if
        !2D yz domain                                                  
        else if (nvygbl > 1 .and. nvzgbl > 1) then
          call DMDACreate2d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                    DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,             &
#else
                    DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                 &
#endif
                    DMDA_STENCIL_BOX,                                  &
                    nvygbl,nvzgbl,PETSC_DECIDE,PETSC_DECIDE,           &
                    dmda_selec%dof, dmda_selec%swidth,                 &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%da,ierr)
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                    dmda_selec%nvygbl,dmda_selec%nvzgbl,               &
                    PETSC_NULL_INTEGER,                                &
                    dmda_selec%npy,dmda_selec%npz,                     &
                    PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,            &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%ys,             &
                    dmda_selec%zs,PETSC_NULL_INTEGER,                  &
                    dmda_selec%yl,dmda_selec%zl,                       &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gys,       &
                    dmda_selec%gzs,PETSC_NULL_INTEGER,                 &
                    dmda_selec%gyl,dmda_selec%gzl,                     &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 2            
            mpiarray_sizes_gbl(1) = nvygbl            
            mpiarray_sizes_sub(1) = dmda_selec%yl
            mpiarray_starts_sub(1) = dmda_selec%ys
            
            mpiarray_sizes_gbl(2) = nvzgbl          
            mpiarray_sizes_sub(2) = dmda_selec%zl
            mpiarray_starts_sub(2) = dmda_selec%zs
            
            mpiarray_sizes_vel_gbl(1) = nvygbl-1 
            if(dmda_selec%ys == dmda_selec%gys) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%yl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%ys
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%yl 
              mpiarray_starts_vel_sub(1) = dmda_selec%ys-1
            end if
            
            mpiarray_sizes_vel_gbl(2) = nvzgbl-1
            if(dmda_selec%zs == dmda_selec%gzs) then
              mpiarray_sizes_vel_sub(2) = dmda_selec%zl-1  
              mpiarray_starts_vel_sub(2) = dmda_selec%zs
            else
              mpiarray_sizes_vel_sub(2) = dmda_selec%zl 
              mpiarray_starts_vel_sub(2) = dmda_selec%zs-1
            end if
          end if
        !2D xz domain                                                  
        else if (nvzgbl > 1 .and. nvxgbl > 1) then  
          call DMDACreate2d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                    DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,             &
#else
                    DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                 &
#endif
                    DMDA_STENCIL_BOX,                                  &
                    nvxgbl,nvzgbl,PETSC_DECIDE,PETSC_DECIDE,           &
                    dmda_selec%dof, dmda_selec%swidth,                 &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%da,ierr)
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                    dmda_selec%nvxgbl,dmda_selec%nvzgbl,               &
                    PETSC_NULL_INTEGER,                                &
                    dmda_selec%npx,dmda_selec%npz,                     &
                    PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,            &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%xs,             &
                    dmda_selec%zs,PETSC_NULL_INTEGER,                  &
                    dmda_selec%xl,dmda_selec%zl,                       &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gxs,       &
                    dmda_selec%gzs,PETSC_NULL_INTEGER,                 &
                    dmda_selec%gxl,dmda_selec%gzl,                     &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 2            
            mpiarray_sizes_gbl(1) = nvxgbl            
            mpiarray_sizes_sub(1) = dmda_selec%xl
            mpiarray_starts_sub(1) = dmda_selec%xs
            
            mpiarray_sizes_gbl(2) = nvzgbl          
            mpiarray_sizes_sub(2) = dmda_selec%zl
            mpiarray_starts_sub(2) = dmda_selec%zs
            
           
            mpiarray_sizes_vel_gbl(1) = nvxgbl-1 
            if(dmda_selec%xs == dmda_selec%gxs) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%xs
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl 
              mpiarray_starts_vel_sub(1) = dmda_selec%xs-1
            end if
            
            mpiarray_sizes_vel_gbl(2) = nvzgbl-1
            if(dmda_selec%zs == dmda_selec%gzs) then
              mpiarray_sizes_vel_sub(2) = dmda_selec%zl-1  
              mpiarray_starts_vel_sub(2) = dmda_selec%zs
            else
              mpiarray_sizes_vel_sub(2) = dmda_selec%zl 
              mpiarray_starts_vel_sub(2) = dmda_selec%zs-1
            end if
          end if
        !1D x domain                                                   
        else if (nvxgbl > 1) then
          call DMDACreate1d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                    DMDA_BOUNDARY_NONE,                                &
#else
                    DM_BOUNDARY_NONE,                                  &
#endif
                    nvxgbl,dmda_selec%dof,dmda_selec%swidth,           &
                    PETSC_NULL_INTEGER,dmda_selec%da,ierr)  
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                    dmda_selec%nvxgbl,PETSC_NULL_INTEGER,              &
                    PETSC_NULL_INTEGER,                                &
                    dmda_selec%npx,PETSC_NULL_INTEGER,                 &
                    PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,            &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%xs,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%xl,PETSC_NULL_INTEGER,                  &
                    PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gxs,       &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%gxl,PETSC_NULL_INTEGER,                 &
                    PETSC_NULL_INTEGER,ierr)  
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 1            
            mpiarray_sizes_gbl(1) = nvxgbl            
            mpiarray_sizes_sub(1) = dmda_selec%xl
            mpiarray_starts_sub(1) = dmda_selec%xs
            
            mpiarray_sizes_vel_gbl(1) = nvxgbl-1 
            if(dmda_selec%xs == dmda_selec%gxs) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%xs
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%xl 
              mpiarray_starts_vel_sub(1) = dmda_selec%xs-1
            end if
          end if                                                            
        !1D y domain                                                   
        else if (nvygbl > 1) then           
          call DMDACreate1d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                    DMDA_BOUNDARY_NONE,                                &
#else
                    DM_BOUNDARY_NONE,                                  &
#endif
                    nvygbl,dmda_selec%dof,dmda_selec%swidth,           &
                    PETSC_NULL_INTEGER,dmda_selec%da,ierr)
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                    dmda_selec%nvygbl,PETSC_NULL_INTEGER,              &
                    PETSC_NULL_INTEGER,                                &
                    dmda_selec%npy,PETSC_NULL_INTEGER,                 &
                    PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,            &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%ys,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%yl,PETSC_NULL_INTEGER,                  &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gys,       &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_selec%gyl,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 1            
            mpiarray_sizes_gbl(1) = nvygbl            
            mpiarray_sizes_sub(1) = dmda_selec%yl
            mpiarray_starts_sub(1) = dmda_selec%ys
            
            mpiarray_sizes_vel_gbl(1) = nvygbl-1 
            if(dmda_selec%ys == dmda_selec%gys) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%yl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%ys
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%yl 
              mpiarray_starts_vel_sub(1) = dmda_selec%ys-1
            end if
          end if                                                             
        !1D z domain                                                   
        else if (nvzgbl > 1) then
          call DMDACreate1d(Petsc_Comm_World,                          &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                    DMDA_BOUNDARY_NONE,                                &
#else
                    DM_BOUNDARY_NONE,                                  &
#endif
                    nvzgbl,dmda_selec%dof,dmda_selec%swidth,           &
                    PETSC_NULL_INTEGER,dmda_selec%da,ierr)
          CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
          call DMSetUp(dmda_selec%da,ierr)
          CHKERRQ(ierr)
#endif
                                                                       
          call DMDAGetInfo(dmda_selec%da,dmda_selec%dim,               &
                    dmda_selec%nvzgbl,PETSC_NULL_INTEGER,              &
                    PETSC_NULL_INTEGER,                                &
                    dmda_selec%npz,PETSC_NULL_INTEGER,                 &
                    PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,            &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetCorners(dmda_selec%da,dmda_selec%zs,             &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%zl,PETSC_NULL_INTEGER,                  &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
                                                                       
          call DMDAGetGhostCorners(dmda_selec%da,dmda_selec%gzs,       &
                    PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,             &
                    dmda_selec%gzl,PETSC_NULL_INTEGER,                 &
                    PETSC_NULL_INTEGER,ierr)
          CHKERRQ(ierr)
            
          !>Note: mpi subbary parameters are 0-based index
          if (bflag_flow) then
            mpiarray_ndim = 1            
            mpiarray_sizes_gbl(1) = nvzgbl            
            mpiarray_sizes_sub(1) = dmda_selec%zl
            mpiarray_starts_sub(1) = dmda_selec%zs
            
            mpiarray_sizes_vel_gbl(1) = nvzgbl-1 
            if(dmda_selec%zs == dmda_selec%gzs) then
              mpiarray_sizes_vel_sub(1) = dmda_selec%zl-1  
              mpiarray_starts_vel_sub(1) = dmda_selec%zs
            else
              mpiarray_sizes_vel_sub(1) = dmda_selec%zl 
              mpiarray_starts_vel_sub(1) = dmda_selec%zs-1
            end if
          end if
        end if

        !> Shift the starting indices up by one so that we can easily
        !> use the Fortran convention of 1-based indices (rather 0-based indices).
        dmda_selec%xs  = dmda_selec%xs+1        
        dmda_selec%ys  = dmda_selec%ys+1        
        dmda_selec%zs  = dmda_selec%zs+1
        
        dmda_selec%gxs = dmda_selec%gxs+1
        dmda_selec%gys = dmda_selec%gys+1
        dmda_selec%gzs = dmda_selec%gzs+1
        
        dmda_selec%xe  = dmda_selec%xs+dmda_selec%xl-1
        dmda_selec%ye  = dmda_selec%ys+dmda_selec%yl-1
        dmda_selec%ze  = dmda_selec%zs+dmda_selec%zl-1

        dmda_selec%gxe = dmda_selec%gxs+dmda_selec%gxl-1
        dmda_selec%gye = dmda_selec%gys+dmda_selec%gyl-1
        dmda_selec%gze = dmda_selec%gzs+dmda_selec%gzl-1

      !c unstructured grid
      else if (discretization_type > 0) then

      end if

      !> Extract global and local vectors from DMDA; then duplicate for remaining
      !> vectors that are the same types
      call DMCreateGlobalVector(dmda_selec%da,x_selec,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(x_selec,b_selec,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(b_selec,r_selec,ierr)
      CHKERRQ(ierr)

      call DMCreateLocalVector(dmda_selec%da,x_loc_selec,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(x_loc_selec,b_loc_selec,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(b_loc_selec,r_loc_selec,ierr)
      CHKERRQ(ierr)

      !> get range
      call VecGetOwnershipRange(x_selec, istart, iend, ierr)
      CHKERRQ(ierr)
      dmda_selec%range_start = istart + 1
      dmda_selec%range_end = iend

      !> Check the result
#ifdef DEBUG
      info_debug = 0
      if(info_debug > 0) then
        if (discretization_type == 0) then
          write(idbg, '(a/,26(1x,a,1x,i4))')                           &
                "flow","dof",dmda_selec%dof,                           &
                "stencil_width", dmda_selec%swidth,                    &
                "dim",dmda_selec%dim,"nvxgbl",dmda_selec%nvxgbl,       &
                "nvygbl",dmda_selec%nvygbl,                            &
                "nvzgbl",dmda_selec%nvzgbl,                            &
                "xs",dmda_selec%xs,"xe",dmda_selec%xe,                 &
                "xl",dmda_selec%xl,"gxs",dmda_selec%gxs,               &
                "gxe",dmda_selec%gxe,"gxl",dmda_selec%gxl,             &
                "ys",dmda_selec%ys,"ye",dmda_selec%ye,                 &
                "yl",dmda_selec%yl,"gys",dmda_selec%gys,               &
                "gye",dmda_selec%gye,"gyl",dmda_selec%gyl,             &
                "zs",dmda_selec%zs,"ze",dmda_selec%ze,                 &
                "zl",dmda_selec%zl,"gzs",dmda_selec%gzs,               &
                "gze",dmda_selec%gze,"gzl",dmda_selec%gzl,             &
                "vec_range_start", dmda_selec%range_start,             &
                "vec_range_end", dmda_selec%range_end
        else if (discretization_type > 0) then
          write(idbg,'(a/,5(a,1x,i6,1x))')                             &
                "flow","dof",dmda_selec%dof,                           &
                "stencil_width",dmda_selec%swidth,                     &
                "dim",dmda_selec%dim,                                  &
                "vec_range_start", dmda_selec%range_start,             &
                "vec_range_end", dmda_selec%range_end
        end if
      end if
#endif
    
    end subroutine solver_dd_DMDACreate_flow_heat
    
    !>
    !> Create distributed array (DMDA) for reactive
    !> transport problem
    !>
    subroutine solver_dd_DMDACreate_react()

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscdmda.h>
      use petscdmda
#endif
      use gen, only: nvxgbl, nvygbl, nvzgbl, n, idbg, rank,            &
                     discretization_type

      implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscdmdef.h>
#include <petsc/finclude/petscdm.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscdmda.h90>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscdmdef.h>
#include <finclude/petscdm.h>
#include <finclude/petscdmda.h>
#include <finclude/petscdmda.h90>
#endif
        
      PetscErrorCode :: ierr
        
      PetscInt :: i, istart, iend, info_debug
       
      !> Initialize
      dmda_react%dim = dmda_flow%dim

      dmda_react%dof = n

      dmda_react%swidth = stencil_width

      !c structured grid
      if (discretization_type == 0) then
        
        dmda_react%xs   = 0
        dmda_react%xe   = 0
        dmda_react%xl   = 1
        dmda_react%gxs  = 0
        dmda_react%gxe  = 0
        dmda_react%gxl  = 1

        dmda_react%ys   = 0
        dmda_react%ye   = 0
        dmda_react%yl   = 1
        dmda_react%gys  = 0
        dmda_react%gye  = 0
        dmda_react%gyl  = 1

        dmda_react%zs   = 0
        dmda_react%ze   = 0
        dmda_react%zl   = 1
        dmda_react%gzs  = 0
        dmda_react%gze  = 0
        dmda_react%gzl  = 1

        dmda_react%nvxgbl   = 1
        dmda_react%nvygbl   = 1
        dmda_react%nvzgbl   = 1

        dmda_react%npx  = 1
        dmda_react%npy  = 1
        dmda_react%npz  = 1
   
        !> Create DMDA and get local grid boundary
        !3D domain
        if (nvxgbl > 1 .and. nvygbl > 1 .and. nvzgbl > 1) then 
!cdsu Note: DMDABoundaryType has become DMBoundaryType, and all the 
!cdsu enumerationvalues have also been renamed from PETSc 3.5.
!cdsu For the windows version, we still use PETSc 3.4. 
            call DMDACreate3d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,            &
                     DMDA_BOUNDARY_NONE,                               &
#else
                     DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                &
                     DM_BOUNDARY_NONE,                                 &
#endif
                     DMDA_STENCIL_BOX,nvxgbl,nvygbl,nvzgbl,            &
                     PETSC_DECIDE,PETSC_DECIDE,                        &
                     PETSC_DECIDE,                                     &
                     dmda_react%dof, dmda_react%swidth,                &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif
                                                                       
            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvxgbl,dmda_react%nvygbl,              &
                     dmda_react%nvzgbl,                                &
                     dmda_react%npx,dmda_react%npy,                    &
                     dmda_react%npz, PETSC_NULL_INTEGER,               &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%xs,           &
                     dmda_react%ys,dmda_react%zs,                      &
                     dmda_react%xl,dmda_react%yl,                      &
                     dmda_react%zl,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gxs,     &
                     dmda_react%gys,dmda_react%gzs,                    &
                     dmda_react%gxl,dmda_react%gyl,                    &
                     dmda_react%gzl,ierr)
             CHKERRQ(ierr)

        !2D xy domain                                                  
        else if (nvxgbl > 1 .and. nvygbl > 1) then  

            call DMDACreate2d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,            &
#else
                     DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                &
#endif
                     DMDA_STENCIL_BOX,                                 &
                     nvxgbl,nvygbl,PETSC_DECIDE,PETSC_DECIDE,          &
                     dmda_react%dof, dmda_react%swidth,                &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif

            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvxgbl,dmda_react%nvygbl,              &
                     PETSC_NULL_INTEGER,                               &
                     dmda_react%npx,dmda_react%npy,                    &
                     PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%xs,           &
                     dmda_react%ys,PETSC_NULL_INTEGER,                 &
                     dmda_react%xl,dmda_react%yl,                      &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gxs,     &
                     dmda_react%gys,PETSC_NULL_INTEGER,                &
                     dmda_react%gxl,dmda_react%gyl,                    &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)

        !2D yz domain                                                  
        else if (nvygbl > 1 .and. nvzgbl > 1) then     

            call DMDACreate2d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,            &
#else
                     DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                &
#endif
                     DMDA_STENCIL_BOX,                                 &
                     nvygbl,nvzgbl,PETSC_DECIDE,PETSC_DECIDE,          &
                     dmda_react%dof, dmda_react%swidth,                &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif

            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvygbl,dmda_react%nvzgbl,              &
                     PETSC_NULL_INTEGER,                               &
                     dmda_react%npy,dmda_react%npz,                    &
                     PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%ys,           &
                     dmda_react%zs,PETSC_NULL_INTEGER,                 &
                     dmda_react%yl,dmda_react%zl,                      &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gys,     &
                     dmda_react%gzs,PETSC_NULL_INTEGER,                &
                     dmda_react%gyl,dmda_react%gzl,                    &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)

        !2D xz domain                                                  
        else if (nvzgbl > 1 .and. nvxgbl > 1) then 
            call DMDACreate2d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,            &
#else
                     DM_BOUNDARY_NONE,DM_BOUNDARY_NONE,                &
#endif
                     DMDA_STENCIL_BOX,                                 &
                     nvxgbl,nvzgbl,PETSC_DECIDE,PETSC_DECIDE,          &
                     dmda_react%dof, dmda_react%swidth,                &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif
                                                                       
            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvxgbl,dmda_react%nvzgbl,              &
                     PETSC_NULL_INTEGER,                               &
                     dmda_react%npx,dmda_react%npz,                    &
                     PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%xs,           &
                     dmda_react%zs,PETSC_NULL_INTEGER,                 &
                     dmda_react%xl,dmda_react%zl,                      &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gxs,     &
                     dmda_react%gzs,PETSC_NULL_INTEGER,                &
                     dmda_react%gxl,dmda_react%gzl,                    &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)

        !1D x domain                                                   
        else if (nvxgbl > 1) then

            call DMDACreate1d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,                               &
#else
                     DM_BOUNDARY_NONE,                                 &
#endif
                     nvxgbl,dmda_react%dof,dmda_react%swidth,          &
                     PETSC_NULL_INTEGER,dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif
                                                                       
            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvxgbl,PETSC_NULL_INTEGER,             &
                     PETSC_NULL_INTEGER,                               &
                     dmda_react%npx,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%xs,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%xl,PETSC_NULL_INTEGER,                 &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gxs,     &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%gxl,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
        !1D y domain                                                   
        else if (nvygbl > 1) then
            call DMDACreate1d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,                               &
#else
                     DM_BOUNDARY_NONE,                                 &
#endif
                     nvygbl,dmda_react%dof,dmda_react%swidth,          &
                     PETSC_NULL_INTEGER,dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif
                                                                       
            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvygbl,PETSC_NULL_INTEGER,             &
                     PETSC_NULL_INTEGER,                               &
                     dmda_react%npy,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%ys,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%yl,PETSC_NULL_INTEGER,                 &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gys,     &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%gyl,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
        !1D z domain                                                   
        else if (nvzgbl > 1) then

            call DMDACreate1d(Petsc_Comm_World,                        &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
                     DMDA_BOUNDARY_NONE,                               &
#else
                     DM_BOUNDARY_NONE,                                 &
#endif
                     nvzgbl,dmda_react%dof,dmda_react%swidth,          &
                     PETSC_NULL_INTEGER,dmda_react%da,ierr)
            CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
            call DMSetUp(dmda_react%da,ierr)
            CHKERRQ(ierr)
#endif
                                                                       
            call DMDAGetInfo(dmda_react%da,dmda_react%dim,             &
                     dmda_react%nvzgbl,PETSC_NULL_INTEGER,             &
                     PETSC_NULL_INTEGER,                               &
                     dmda_react%npz,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER, PETSC_NULL_INTEGER,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetCorners(dmda_react%da,dmda_react%zs,           &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%zl,PETSC_NULL_INTEGER,                 &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
                                                                       
            call DMDAGetGhostCorners(dmda_react%da,dmda_react%gzs,     &
                     PETSC_NULL_INTEGER,PETSC_NULL_INTEGER,            &
                     dmda_react%gzl,PETSC_NULL_INTEGER,                &
                     PETSC_NULL_INTEGER,ierr)
            CHKERRQ(ierr)
            
        end if
        
        !> Shift the starting indices up by one so that we can easily
        !> use the Fortran convention of 1-based indices (rather 0-based indices).
        dmda_react%xs  = dmda_react%xs+1
        dmda_react%ys  = dmda_react%ys+1
        dmda_react%zs  = dmda_react%zs+1

        dmda_react%gxs = dmda_react%gxs+1
        dmda_react%gys = dmda_react%gys+1
        dmda_react%gzs = dmda_react%gzs+1

        dmda_react%xe  = dmda_react%xs+dmda_react%xl-1
        dmda_react%ye  = dmda_react%ys+dmda_react%yl-1
        dmda_react%ze  = dmda_react%zs+dmda_react%zl-1

        dmda_react%gxe = dmda_react%gxs+dmda_react%gxl-1
        dmda_react%gye = dmda_react%gys+dmda_react%gyl-1
        dmda_react%gze = dmda_react%gzs+dmda_react%gzl-1
        
      !c unstructured grid
      else if (discretization_type > 0) then

      end if

      !> Extract global and local vectors from DMDA; then duplicate for remaining
      !> vectors that are the same types
      call DMCreateGlobalVector(dmda_react%da,x_react,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(x_react,b_react,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(b_react,r_react,ierr)
      CHKERRQ(ierr)

      call DMCreateLocalVector(dmda_react%da,x_react_loc,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(x_react_loc,b_react_loc,ierr)
      CHKERRQ(ierr)
      call VecDuplicate(b_react_loc,r_react_loc,ierr)
      CHKERRQ(ierr)

      !> get range
      call VecGetOwnershipRange(x_react, istart, iend, ierr)
      CHKERRQ(ierr)
      dmda_react%range_start = istart + 1
      dmda_react%range_end = iend

      !> Check the result
#ifdef DEBUG
      info_debug = 0
      if(info_debug > 0) then
        if (discretization_type == 0) then
            write(idbg, '(a/,26(1x,a,1x,i4))')                         &
                  "reactive transport","dof",dmda_react%dof,           &
                  "stencil_width", dmda_react%swidth,                  &
                  "dim",dmda_react%dim,"nvxgbl",dmda_react%nvxgbl,     &
                  "nvygbl",dmda_react%nvygbl,                          &
                  "nvzgbl",dmda_react%nvzgbl,                          &
                  "xs",dmda_react%xs,"xe",dmda_react%xe,               &
                  "xl",dmda_react%xl,"gxs",dmda_react%gxs,             &
                  "gxe",dmda_react%gxe,"gxl",dmda_react%gxl,           &
                  "ys",dmda_react%ys,"ye",dmda_react%ye,               &
                  "yl",dmda_react%yl,"gys",dmda_react%gys,             &
                  "gye",dmda_react%gye,"gyl",dmda_react%gyl,           &
                  "zs",dmda_react%zs,"ze",dmda_react%ze,               &
                  "zl",dmda_react%zl,"gzs",dmda_react%gzs,             &
                  "gze",dmda_react%gze,"gzl",dmda_react%gzl,           &
                  "vec_range_start", dmda_react%range_start,           &
                  "vec_range_end", dmda_react%range_end
        else if (discretization_type > 0) then
          write(idbg,'(a/,5(a,1x,i6,1x))')                             &
                "react","dof",dmda_react%dof,                          &
                "stencil_width",dmda_react%swidth,                     &
                "dim",dmda_react%dim,                                  &
                "vec_range_start", dmda_react%range_start,             &
                "vec_range_end", dmda_react%range_end
        end if
      end if
#endif
    
    end subroutine solver_dd_DMDACreate_react
    
    !>Set global and local node index map
    subroutine solver_dd_mapping_set
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscis.h>
#include <petsc/finclude/petscdmda.h>
        use petscis
        use petscdmda
#endif
        use gen, only: idbg, nvxgbl, nvygbl, nvzgbl, nvxls, nvxle,     &
                       nvyls, nvyle, nvzls, nvzle, nvxgls, nvxgle,     &
                       nvygls, nvygle, nvzgls, nvzgle, nn, nngl,       &
                       nngbl, node_idx_lg2l, node_idx_lg2g,            &
                       node_idx_lg2pg, node_idx_l2lg, rank, ilog,      &
                       mem_cur, mem_max, memory_monitor,               &
                       node_idx_vel_lg2g
        use petsc_mpi_common, only : petsc_mpi_finalize
    
        implicit none

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscis.h>
#include <petsc/finclude/petscdmdef.h>
#include <petsc/finclude/petscdm.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscdmda.h90>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscis.h>
#include <finclude/petscdmdef.h>
#include <finclude/petscdm.h>
#include <finclude/petscdmda.h>
#include <finclude/petscdmda.h90>
#endif

        
        PetscErrorCode :: ierr
               
        PetscInt :: ivx, ivy, ivz, ivol, ivol_l, ibase, ig, dof, ig_vel 

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
        PetscInt, pointer :: idx(:)
#else
        ISLocalToGlobalMapping :: ltogm
        PetscInt :: ltog(1)
        PetscOffset :: idltog
#endif
        
        PetscInt :: info_debug
        
        external :: checkerr

        
        allocate(node_idx_lg2g(nngl), stat = ierr)
        call checkerr(ierr,'node_idx_lg2g',ilog)
        node_idx_lg2g = 0
        call memory_monitor(sizeof(node_idx_lg2g),'node_idx_lg2g',.true.)
        
        allocate(node_idx_vel_lg2g(nngl), stat = ierr)
        call checkerr(ierr,'node_idx_vel_lg2g',ilog)
        node_idx_vel_lg2g = 0
        call memory_monitor(sizeof(node_idx_vel_lg2g),'node_idx_vel_lg2g',.true.)
        
        allocate(node_idx_lg2l(nngl), stat = ierr)
        call checkerr(ierr,'node_idx_lg2l',ilog)
        node_idx_lg2l = 0
        call memory_monitor(sizeof(node_idx_lg2l),'node_idx_lg2l',.true.)
        
        allocate(node_idx_lg2pg(nngl), stat = ierr)
        call checkerr(ierr,'node_idx_lg2pg',ilog)
        node_idx_lg2pg = 0
        call memory_monitor(sizeof(node_idx_lg2pg),'node_idx_lg2pg',.true.)
        
        allocate(node_idx_l2lg(nn), stat = ierr)
        call checkerr(ierr,'node_idx_l2lg',ilog)
        node_idx_l2lg = 0

        !allocate(node_idx_lb2pg(nngl), stat = ierr)
        !call checkerr(ierr,'node_idx_lb2pg',ilog)
        !node_idx_lb2pg = -1

        ivol = 1
        ivol_l = 1
        
        select case (dmda_flow%dim)
            
        case (1)
            
            !1D x domain    
            if (nvxgbl > 1) then 
                do ivx = nvxgls,nvxgle  
                    node_idx_lg2g(ivol) = ivx
                    if (ivx == 1) then
                      node_idx_vel_lg2g(ivol) = -1
                    else
                      node_idx_vel_lg2g(ivol) = ivx-1 
                    end if
                    if(ivx < nvxls .or. ivx > nvxle) then
                        node_idx_lg2l(ivol) = -1
                    else
                        node_idx_lg2l(ivol) = ivol_l
                        node_idx_l2lg(ivol_l) = ivol
                        ivol_l = ivol_l + 1
                    end if
                    !if (ivx == nvxls .and. ivx > nvxgls) then
                    !  node_idx_lb2pg(ivol) = 1
                    !end if
                    ivol = ivol + 1
                end do
            !1D y domain
            else if (nvygbl > 1) then  
                do ivy = nvygls,nvygle
                    node_idx_lg2g(ivol) = ivy
                    if (ivy == 1) then
                      node_idx_vel_lg2g(ivol) = -1
                    else
                      node_idx_vel_lg2g(ivol) = ivy-1 
                    end if
                    if(ivy < nvyls .or. ivy > nvyle) then
                        node_idx_lg2l(ivol) = -1
                    else
                        node_idx_lg2l(ivol) = ivol_l
                        node_idx_l2lg(ivol_l) = ivol
                        ivol_l = ivol_l + 1
                    end if
                    !if (ivy == nvyls .and. ivy > nvygls) then
                    !  node_idx_lb2pg(ivol) = 1
                    !end if
                    ivol = ivol + 1
                end do
            !1D z domain
            else if (nvzgbl > 1) then 
                do ivz = nvzgls, nvzgle
                    node_idx_lg2g(ivol) = ivz
                    if (ivz == 1) then
                      node_idx_vel_lg2g(ivol) = -1
                    else
                      node_idx_vel_lg2g(ivol) = ivz-1 
                    end if
                    if(ivz < nvzls .or. ivz > nvzle) then
                        node_idx_lg2l(ivol) = -1
                    else
                        node_idx_lg2l(ivol) = ivol_l
                        node_idx_l2lg(ivol_l) = ivol
                        ivol_l = ivol_l + 1
                    end if
                    !if (ivz == nvzls .and. ivz > nvzgls) then
                    !  node_idx_lb2pg(ivol) = 1
                    !end if
                    ivol = ivol + 1
                end do  
            end if              
            
        case (2)
           
            !2D xy domain    
            if (nvxgbl > 1 .and. nvygbl > 1) then
                do ivy = nvygls,nvygle
                    ig = (ivy-1)*nvxgbl
                    ig_vel = (ivy-2)*(nvxgbl-1)
                    do ivx = nvxgls,nvxgle
                        node_idx_lg2g(ivol) = ig + ivx
                        if (ivx == 1 .or. ivy == 1) then
                          node_idx_vel_lg2g(ivol) = -1
                        else
                          node_idx_vel_lg2g(ivol) = ig_vel+ivx-1 
                        end if
                        if(ivx < nvxls .or. ivx > nvxle .or.           &
                           ivy < nvyls .or. ivy > nvyle) then
                            node_idx_lg2l(ivol) = -1
                        else
                            node_idx_lg2l(ivol) = ivol_l
                            node_idx_l2lg(ivol_l) = ivol
                            ivol_l = ivol_l + 1     
                        end if
                        !if ((ivx == nvxls .and. ivx > nvxgls) .or.     &
                        !    (ivy == nvyls .and. ivy > nvygls)) then
                        !  node_idx_lb2pg(ivol) = 1
                        !end if
                        ivol = ivol + 1   
                    end do
                end do
            !2D yz domain    
            else if (nvygbl > 1 .and. nvzgbl > 1) then
                do ivz = nvzgls, nvzgle
                    ig = (ivz-1)*nvygbl
                    ig_vel = (ivz-2)*(nvygbl-1)
                    do ivy = nvygls,nvygle            
                        node_idx_lg2g(ivol) = ig + ivy
                        if (ivz == 1 .or. ivy == 1) then
                          node_idx_vel_lg2g(ivol) = -1
                        else
                          node_idx_vel_lg2g(ivol) = ig_vel+ivy-1 
                        end if
                        if(ivy < nvyls .or. ivy > nvyle .or.           &
                           ivz < nvzls .or. ivz > nvzle) then
                            node_idx_lg2l(ivol) = -1
                        else    
                            node_idx_lg2l(ivol) = ivol_l
                            node_idx_l2lg(ivol_l) = ivol
                            ivol_l = ivol_l + 1
                        end if
                        !if ((ivy == nvyls .and. ivy > nvygls) .or.     &
                        !    (ivz == nvzls .and. ivz > nvzgls)) then
                        !  node_idx_lb2pg(ivol) = 1
                        !end if
                        ivol = ivol + 1     
                    end do
                end do
            !2D xz domain    
            else if (nvzgbl > 1 .and. nvxgbl > 1) then
                do ivz = nvzgls, nvzgle
                    ig = (ivz-1)*nvxgbl
                    ig_vel = (ivz-2)*(nvxgbl-1)
                    do ivx = nvxgls,nvxgle                   
                        node_idx_lg2g(ivol) = ig + ivx
                        if (ivz == 1 .or. ivx == 1) then
                          node_idx_vel_lg2g(ivol) = -1
                        else
                          node_idx_vel_lg2g(ivol) = ig_vel+ivx-1 
                        end if
                        if(ivx < nvxls .or. ivx > nvxle .or.           &
                           ivz < nvzls .or. ivz > nvzle) then
                            node_idx_lg2l(ivol) = -1
                        else    
                            node_idx_lg2l(ivol) = ivol_l
                            node_idx_l2lg(ivol_l) = ivol
                            ivol_l = ivol_l + 1
                        end if
                        !if ((ivx == nvxls .and. ivx > nvxgls) .or.     &
                        !    (ivz == nvzls .and. ivz > nvzgls)) then
                        !  node_idx_lb2pg(ivol) = 1
                        !end if
                        ivol = ivol + 1
                    end do
                end do
            end if
            
        case (3)             

            do ivz = nvzgls, nvzgle                
                do ivy = nvygls,nvygle
                    ig = (ivz-1)*nvygbl*nvxgbl+(ivy-1)*nvxgbl
                    ig_vel = (ivz-2)*(nvygbl-1)*(nvxgbl-1)+            &
                             (ivy-2)*(nvxgbl-1)
                    do ivx = nvxgls,nvxgle   
                        node_idx_lg2g(ivol) = ig + ivx
                        if (ivx == 1 .or. ivy == 1 .or. ivz == 1) then
                          node_idx_vel_lg2g(ivol) = -1
                        else
                          node_idx_vel_lg2g(ivol) = ig_vel+ivx-1 
                        end if
                        if(ivx < nvxls .or. ivx > nvxle .or.           &
                           ivy < nvyls .or. ivy > nvyle .or.           &
                           ivz < nvzls .or. ivz > nvzle) then
                            node_idx_lg2l(ivol) = -1
                        else
                            node_idx_lg2l(ivol) = ivol_l
                            node_idx_l2lg(ivol_l) = ivol
                            ivol_l = ivol_l + 1    
                        end if
                        !if ((ivx == nvxls .and. ivx > nvxgls) .or.     &
                        !    (ivy == nvyls .and. ivy > nvygls) .or.     &
                        !    (ivz == nvzls .and. ivz > nvzgls)) then
                        !  node_idx_lb2pg(ivol) = 1
                        !end if

                        ivol = ivol + 1                
                    end do
                end do
            end do
          
        end select
        
        info_debug = 0
#ifdef DEBUG
        if(info_debug > 0) then
            do ivol = 1, nngl
                write(idbg,'(5(a,1x,i6,1x))') "ivol_l",ivol,           &
                      "ivol_l_no_ghost",node_idx_lg2l(ivol),           &
                      "ivol_g",node_idx_lg2g(ivol),                    &
                      "ivol_vel_g",node_idx_vel_lg2g(ivol)
                      
            end do
        end if
#endif        
        
        !Check DMDA pointer to the list of global indices in the matrix
        !When dof == 2, size(idx,1) = 2*nngl
!cdsu Note:  DMDAGetGlobalIndicesF90 has been removed from PETSc 3.5, 
!cdsu Note:  use DMGetLocalToGlobalMapping instead.
!cdsu For the windows version, we still use PETSc 3.4. 


#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
        call DMDAGetGlobalIndicesF90(dmda_flow%da,PETSC_NULL_INTEGER,  &
                                     idx,ierr)
        CHKERRQ(ierr)
#else
        call DMGetLocalToGlobalMapping(dmda_flow%da,ltogm,ierr)
        CHKERRQ(ierr)
        call ISLocalToGlobalMappingGetIndices(ltogm,ltog,idltog,ierr)
        CHKERRQ(ierr)
#endif
        
        dof = dmda_flow%dof
      
        do ivol = 1, nngl

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)
            node_idx_lg2pg(ivol) = (idx(ivol*dof)+1)/dof
#else
            node_idx_lg2pg(ivol) = (ltog(ivol*dof + idltog)+1)/dof
#endif


#ifdef DEBUG
            if(info_debug > 0) then
              write(idbg,'(2(a,1x,i10,1x))')"ivol",ivol,"global index",&
                                          node_idx_lg2pg(ivol)
            end if
#endif
        end do

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR <= 4)

#else
        call ISLocalToGlobalMappingRestoreIndices(ltogm,ltog,idltog,ierr)
        CHKERRQ(ierr)
#endif
        
        if(info_debug > 1) then
            call petsc_mpi_finalize
            stop
        end if
        
   
    end subroutine solver_dd_mapping_set
   
    !> set DMDA coordinates 
    subroutine solver_dd_coordinates_set
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)

#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscdmda.h>
        use petscvec
        use petscdmda
#endif
        use gen, only: idbg, nvxgbl, nvygbl, nvzgbl, nvxls, nvxle,     &
                       nvyls, nvyle, nvzls, nvzle, nvxgls, nvxgle,     &
                       nvygls, nvygle, nvzgls, nvzgle,                 &
                       nvxgl, nvygl, nvzgl, xglat, yglat, zglat
    
        implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscdmdef.h>
#include <petsc/finclude/petscdm.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscdmda.h90>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscdmdef.h>
#include <finclude/petscdm.h>
#include <finclude/petscdmda.h>
#include <finclude/petscdmda.h90>
#endif
        
        PetscErrorCode :: ierr
        
        PetscScalar, pointer :: coords1d(:)
        !DMDACoor2d, pointer :: coords2d(:,:)
        !DMDACoor3d, pointer :: coords3d(:,:,:) 
        !Note: There is no DMDACoor2d or 3d in Fortran (maybe added later), 
        !should use PetscScalar instead. 
        !The first index is 0, 1 or 2 for the x, y, or z coordinate
        PetscScalar, pointer :: coords2d(:,:,:) 
        PetscScalar, pointer :: coords3d(:,:,:,:) 
        PetscReal :: xmin, xmax, ymin, ymax, zmin, zmax
        Vec :: gc,global
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 11)
        type(tDM) :: cda
#else
        DM :: cda
#endif
        PetscInt :: ivx, ivy, ivz, ivol, istart1, iend1,               &
                    istart2, iend2, istart3, iend3, istart4, iend4,    &
                    ibase        

        PetscReal :: r0
        
        
        PetscInt :: info_debug
        
        parameter (r0 = 0.0d0)
        
        info_debug = 0

        xmin = 0.0
        xmax = 0.0
        ymin = 0.0
        ymax = 0.0
        zmin = 0.0
        zmax = 0.0
      
        
        select case (dmda_flow%dim)
            
        case (1)
            
            xmax = 1.0
            call DMDASetUniformCoordinates(dmda_flow%da,               &
                    xmin, xmax, ymin, ymax, zmin, zmax, ierr)
            CHKERRQ(ierr)
            
            call DMGetCoordinateDM(dmda_flow%da,cda,ierr)
            CHKERRQ(ierr)
            call DMGetCoordinatesLocal(dmda_flow%da,gc,ierr)
            CHKERRQ(ierr)
            call DMDAVecGetArrayF90(cda,gc,coords1d,ierr)
            CHKERRQ(ierr)
     
#ifdef DEBUG
            if(info_debug > 0) then
                !Please note the index is 0-based            
                istart1 = lbound(coords1d,1)
                iend1 = ubound(coords1d,1)  
                write(idbg,'(a/,2(a,1x,i10,1x))')                      &
                        "set the coordinates 1 dim",                   &
                        "start1",istart1,"end1",iend1
            end if
#endif
            
            ibase = 1 
            ivol = 0
            
            !1D x domain
            if (nvxgbl > 1) then                                
                do ivx = nvxls,nvxle
                    coords1d(ivx-ibase) = xglat(ivx)   
#ifdef DEBUG
                    if(info_debug > 0) then
                        ivol = ivx - nvxgls + 1
                        write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))')   &
                                "ivol_l", ivol,                        &
                                "xg", coords1d(ivx-ibase),             &
                                "yg", r0,                              &
                                "zg", r0
                    end if
#endif
                end do
            !1D y domain
            else if (nvygbl > 1) then                                
                do ivy = nvyls,nvyle
                    coords1d(ivy-ibase) = yglat(ivy)
#ifdef DEBUG
                    if(info_debug > 0) then
                        ivol = ivy - nvygls + 1
                        write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))')   &
                                "ivol_l", ivol,                        &
                                "xg", r0,                              &
                                "yg", coords1d(ivy-ibase),             &
                                "zg", r0
                    end if
#endif
                end do
            !1D z domain
            else if (nvzgbl > 1) then 
                do ivz = nvzls, nvzle
                    coords1d(ivz-ibase) = zglat(ivz)
#ifdef DEBUG
                    if(info_debug > 0) then
                        ivol = ivz - nvzgls + 1
                        write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))')   &
                                "ivol_l", ivol,                        &
                                "xg", r0,                              &
                                "yg", r0,                              &
                                "zg", coords1d(ivz-ibase)
                    end if
#endif
                end do  
            end if            

            call DMDAVecRestoreArrayF90(cda,gc,coords1d,ierr)
            CHKERRQ(ierr)
            
        case (2)
            xmax = 1.0
            ymax = 1.0
            call DMDASetUniformCoordinates(dmda_flow%da,               &
                    xmin, xmax, ymin, ymax, zmin, zmax, ierr)
            CHKERRQ(ierr)
            call DMGetCoordinateDM(dmda_flow%da,cda,ierr)
            CHKERRQ(ierr)
            call DMGetCoordinatesLocal(dmda_flow%da,gc,ierr)
            CHKERRQ(ierr)
            call DMDAVecGetArrayF90(cda,gc,coords2d,ierr)
            CHKERRQ(ierr)
            
            !Please note the index is 0-based
#ifdef DEBUG
            if(info_debug > 0) then
                istart1 = lbound(coords2d,1)
                iend1 = ubound(coords2d,1)  
            
                istart2 = lbound(coords2d,2)
                iend2 = ubound(coords2d,2) 
            
                istart3 = lbound(coords2d,3)
                iend3 = ubound(coords2d,3)
            
                write(idbg,'(a/,6(a,1x,i10,1x))')                      &
                      "set the coordinates 2 dim",                     &
                      "start1",istart1,"end1",iend1,                   &
                      "start2",istart2,"end2",iend2,                   &
                      "start3",istart3,"end3",iend3  
            end if
#endif
                                                                       
            ibase = 1 
            ivol = 1
                                                                       
            !2D xy domain                                              
            if (nvxgbl > 1 .and. nvygbl > 1) then                      
                do ivy = nvyls,nvyle                                   
                    do ivx = nvxls,nvxle                               
                        coords2d(1-ibase,ivx-ibase,ivy-ibase)          &
                            = xglat(ivx)                                 
                        coords2d(2-ibase,ivx-ibase,ivy-ibase)          &
                            = yglat(ivy)
#ifdef DEBUG
                        if(info_debug > 0) then
                          ivol = (ivy-nvygls)*nvxgl + ivx - nvxgls + 1   
                          write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))') &
                               "ivol_l", ivol,                         &
                               "xg", coords2d(1-ibase,ivx-ibase,       &
                                              ivy-ibase),              &
                               "yg", coords2d(2-ibase,ivx-ibase,       &
                                              ivy-ibase),              &
                               "zg", r0 
                        end if
#endif
                    end do
                end do
            !2D yz domain    
            else if (nvygbl > 1 .and. nvzgbl > 1) then                  
                do ivz = nvzls, nvzle
                    do ivy = nvyls,nvyle
                        coords2d(1-ibase,ivy-ibase,ivz-ibase)          &
                            = yglat(ivy)                               
                        coords2d(2-ibase,ivy-ibase,ivz-ibase)          &
                            = zglat(ivz)   
#ifdef DEBUG
                        if(info_debug > 0) then
                          ivol = (ivz-nvzgls)*nvygl + ivy-nvygls+1   
                          write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))') &
                               "ivol_l", ivol,                         &
                               "xg", r0,                               &
                               "yg", coords2d(1-ibase,ivy-ibase,       &
                                              ivz-ibase),              &
                               "zg", coords2d(2-ibase,ivy-ibase,       &
                                              ivz-ibase) 
                        end if
#endif
                    end do                                             
                end do                                                 
            !2D xz domain                                              
            else if (nvzgbl > 1 .and. nvxgbl > 1) then                   
                do ivz = nvzls, nvzle                                  
                    do ivx = nvxls,nvxle                               
                        coords2d(1-ibase,ivx-ibase,ivz-ibase)          &
                            = xglat(ivx)                               
                        coords2d(2-ibase,ivx-ibase,ivz-ibase)          &
                            = zglat(ivz)
#ifdef DEBUG
                        if(info_debug > 0) then
                          ivol = (ivz-nvzgls)*nvxgl + ivx-nvxgls+1  
                          write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))') &
                               "ivol_l", ivol,                         &
                               "xg", coords2d(1-ibase,ivx-ibase,       &
                                              ivz-ibase),              &
                               "yg", r0,                               &
                               "zg", coords2d(2-ibase,ivx-ibase,       &
                                              ivz-ibase)
                        end if
#endif
                    end do
                end do
                
            end if            

            call DMDAVecRestoreArrayF90(cda,gc,coords2d,ierr)
            CHKERRQ(ierr)
            
        case (3) 
            xmax = 1.0
            ymax = 1.0
            zmax = 1.0
            call DMDASetUniformCoordinates(dmda_flow%da,               &
                    xmin, xmax, ymin, ymax, zmin, zmax, ierr)
            CHKERRQ(ierr)
            call DMGetCoordinateDM(dmda_flow%da,cda,ierr)
            CHKERRQ(ierr)
            call DMGetCoordinatesLocal(dmda_flow%da,gc,ierr)
            CHKERRQ(ierr)
            call DMDAVecGetArrayF90(cda,gc,coords3d,ierr)
            CHKERRQ(ierr)
        
            !Please note the index is 0-based
#ifdef DEBUG
            if(info_debug > 0) then
                istart1 = lbound(coords3d,1)
                iend1 = ubound(coords3d,1)  
            
                istart2 = lbound(coords3d,2)
                iend2 = ubound(coords3d,2) 
            
                istart3 = lbound(coords3d,3)
                iend3 = ubound(coords3d,3) 
            
                istart4 = lbound(coords3d,4)
                iend4 = ubound(coords3d,4) 
        
                write(idbg,'(a/,8(a,1x,i10,1x))')                      &
                      "set the coordinates 3 dim",                     &
                      "start1",istart1,"end1",iend1,                   &
                      "start2",istart2,"end2",iend2,                   &
                      "start3",istart3,"end3",iend3,                   &
                      "start4",istart4,"end4",iend4
            end if
#endif
        
            ibase = 1
            do ivz = nvzls,nvzle
                do ivy = nvyls,nvyle
                    do ivx = nvxls,nvxle
                        coords3d(1-ibase,ivx-ibase,ivy-ibase,ivz-ibase)& 
                            = xglat(ivx)
                        coords3d(2-ibase,ivx-ibase,ivy-ibase,ivz-ibase)&
                            = yglat(ivy)
                        coords3d(3-ibase,ivx-ibase,ivy-ibase,ivz-ibase)&
                            = zglat(ivz)
#ifdef DEBUG
                        if(info_debug > 0) then
                          ivol = (ivz-nvzgls)*nvygl*nvxgl +            &
                                 (ivy-nvygls)*nvxgl + ivx-nvxgls+1   
                          write(idbg,'(a,1x,i6,3(1x,a,1x,1pe15.6e3))') &
                                  "ivol_l", ivol,                      &
                                  "xg", coords3d(1-ibase,ivx-ibase,    &
                                           ivy-ibase,ivz-ibase),       &
                                  "yg", coords3d(2-ibase,ivx-ibase,    &
                                           ivy-ibase,ivz-ibase),       &
                                  "zg", coords3d(3-ibase,ivx-ibase,    &
                                           ivy-ibase,ivz-ibase)
                        end if
#endif
                    end do
                end do
            end do
            
            call DMDAVecRestoreArrayF90(cda,gc,coords3d,ierr)
            CHKERRQ(ierr)
            
        end select
        
        call DMGetCoordinates(dmda_flow%da,global,ierr)
        CHKERRQ(ierr)
        call DMLocalToGlobalBegin(cda,gc,INSERT_VALUES,global,ierr)
        CHKERRQ(ierr)
        call DMLocalToGlobalEnd(cda,gc,INSERT_VALUES,global,ierr)
        CHKERRQ(ierr)
        call DMRestoreLocalVector(cda,gc,ierr)
        CHKERRQ(ierr)

    end subroutine solver_dd_coordinates_set
    
    
    !> get DMDA coordinates
    subroutine solver_dd_coordinates_get

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscdmda.h>
        use petscvec
        use petscdmda
#endif
        use gen, only: idbg, nvxgbl, nvygbl, nvzgbl, nvxls, nvxle,     &
                       nvyls, nvyle, nvzls, nvzle, nvxgls, nvxgle,     &
                       nvygls, nvygle, nvzgls, nvzgle, nvxgl, nvygl,   &
                       nvzgl, node_idx_lg2g, rank
    
        implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscdmdef.h>
#include <petsc/finclude/petscdm.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscdmda.h90>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscdmdef.h>
#include <finclude/petscdm.h>
#include <finclude/petscdmda.h>
#include <finclude/petscdmda.h90>
#endif
        
        PetscErrorCode :: ierr
        
        PetscScalar, pointer :: coords1d(:)
        !DMDACoor2d, pointer :: coords2d(:,:)
        !DMDACoor3d, pointer :: coords3d(:,:,:) 
        !Note: There is no DMDACoor2d or 3d in Fortran (maybe added later), 
        !should use PetscScalar instead. 
        !The first index is 0, 1 or 2 for the x, y, or z coordinate
        PetscScalar, pointer :: coords2d(:,:,:) 
        PetscScalar, pointer :: coords3d(:,:,:,:)   
        Vec :: gc
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 11)
        type(tDM) :: cda
#else
        DM :: cda
#endif
        PetscInt :: i, ivol, ivx, ivy, ivz, istart1, iend1,            &
                    istart2, iend2, istart3, iend3, istart4, iend4,    &
                    ibase
        PetscReal :: r0
        PetscReal :: lmin(3), lmax(3)
        PetscReal :: gmin(3), gmax(3)        
       
        PetscInt :: info_debug
        
        parameter (r0 = 0.0d0)
        
        info_debug = 0
        
        !Initialize coordinates for bounding box
        dmda_flow%xlmin = 0.0d0
        dmda_flow%xlmax = 0.0d0
        dmda_flow%ylmin = 0.0d0
        dmda_flow%ylmax = 0.0d0
        dmda_flow%zlmin = 0.0d0
        dmda_flow%zlmax = 0.0d0
        dmda_flow%xlmingbl = 0.0d0
        dmda_flow%xlmaxgbl = 0.0d0
        dmda_flow%ylmingbl = 0.0d0
        dmda_flow%ylmaxgbl = 0.0d0
        dmda_flow%zlmingbl = 0.0d0
        dmda_flow%zlmaxgbl = 0.0d0        
        
        !Check coordinates
        call DMGetCoordinateDM(dmda_flow%da,cda,ierr)
        CHKERRQ(ierr)
        call DMGetCoordinates(dmda_flow%da,gc,ierr)
        CHKERRQ(ierr)

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 12)
        call DMGetLocalBoundingBox(dmda_flow%da,lmin,lmax,ierr)
        CHKERRQ(ierr)
        call DMGetBoundingBox(dmda_flow%da,gmin,gmax,ierr)
        CHKERRQ(ierr)
#else
        call DMDAGetLocalBoundingBox(dmda_flow%da,lmin,lmax,ierr)
        CHKERRQ(ierr)
        call DMDAGetBoundingBox(dmda_flow%da,gmin,gmax,ierr)
        CHKERRQ(ierr)
#endif
        select case (dmda_flow%dim)
            
        case (1)
#ifdef DEBUG
            if(info_debug > 0) then
                write(idbg,'(2(/a/,2(a,1x,1pe15.6e3,1x)))')            &
                      "local bounding box 1 dim",                      &
                      "lmin",lmin(1),"lmax",lmax(1),                   &
                      "global bounding box 1 dim",                     & 
                      "gmin",gmin(1),"gmax",gmax(1)
            end if
#endif
            
            call DMDAVecGetArrayF90(cda,gc,coords1d,ierr)
            CHKERRQ(ierr)
        
            !Please note the index is 0-based
#ifdef DEBUG
            if(info_debug > 0) then
                istart1 = lbound(coords1d,1)
                iend1 = ubound(coords1d,1)        
        
                write(idbg,'(a/,2(a,1x,i10,1x))')                      &
                      "get the coordinates 1 dim",                     &
                      "start1",istart1,"end1",iend1
            end if
#endif
            
            ibase = 1 
            ivol = 0
            
            !1D x domain    
            if (nvxgbl > 1) then 
                dmda_flow%xlmin=lmin(1)
                dmda_flow%xlmax=lmax(1)
                
                dmda_flow%xlmingbl=gmin(1)
                dmda_flow%xlmaxgbl=gmax(1)
#ifdef DEBUG
                if(info_debug > 0) then
                  do ivx = nvxls,nvxle
                    ivol = ivx - nvxgls + 1
                    write(idbg,'(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))') &
                          "ivol_l", ivol,                              &
                          "ivol_g", node_idx_lg2g(ivol),               &
                          "xg", coords1d(ivx-ibase),                   &
                          "yg", r0,                                    &
                          "zg", r0
                  end do
                end if
#endif
            !1D y domain
            else if (nvygbl > 1) then  
                dmda_flow%ylmin=lmin(1)
                dmda_flow%ylmax=lmax(1)
                
                dmda_flow%ylmingbl=gmin(1)
                dmda_flow%ylmaxgbl=gmax(1)
#ifdef DEBUG
                if(info_debug > 0) then
                  do ivy = nvyls,nvyle
                    ivol = ivy - nvygls + 1
                    write(idbg,'(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))') &
                          "ivol_l", ivol,                              &
                          "ivol_g", node_idx_lg2g(ivol),               &
                          "xg", r0,                                    &
                          "yg", coords1d(ivy-ibase),                   &
                          "zg", r0
                  end do
                end if
#endif
            !1D z domain
            else if (nvzgbl > 1) then 
                dmda_flow%zlmin=lmin(1)
                dmda_flow%zlmax=lmax(1)
                
                dmda_flow%zlmingbl=gmin(1)
                dmda_flow%zlmaxgbl=gmax(1)
#ifdef DEBUG
                if(info_debug > 0) then
                  do ivz = nvzls, nvzle
                    ivol = ivz - nvzgls + 1
                    write(idbg,'(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))') &
                          "ivol_l", ivol,                              &
                          "ivol_g", node_idx_lg2g(ivol),               &
                          "xg", r0,                                    &
                          "yg", r0,                                    &
                          "zg", coords1d(ivz-ibase)
                  end do
                end if
#endif
            end if              
        
            call DMDAVecRestoreArrayF90(cda,gc,coords1d,ierr)
            CHKERRQ(ierr)
            
        case (2)
#ifdef DEBUG
            if(info_debug > 0) then
                write(idbg,'(2(/a/,4(a,1x,1pe15.6e3,1x)))')            &
                      "local bounding box 2 dim",                      &
                      "lmin",lmin(1),"lmax",lmax(1),                   &
                      "lmin",lmin(2),"lmax",lmax(2),                   &
                      "global bounding box 2 dim",                     &
                      "gmin",gmin(1),"gmax",gmax(1),                   &
                      "gmin",gmin(2),"gmax",gmax(2)
            end if
#endif
            
            call DMDAVecGetArrayF90(cda,gc,coords2d,ierr)
            CHKERRQ(ierr)
        
            !Please note the index is 0-based
#ifdef DEBUG
            if(info_debug > 0) then
                istart1 = lbound(coords2d,1)
                iend1 = ubound(coords2d,1)  
            
                istart2 = lbound(coords2d,2)
                iend2 = ubound(coords2d,2)
            
                istart3 = lbound(coords2d,3)
                iend3 = ubound(coords2d,3)
        
                write(idbg,'(a/,6(a,1x,i10,1x))')                      &
                      "get the coordinates 2 dim",                     &
                      "start1",istart1,"end1",iend1,                   &
                      "start2",istart2,"end2",iend2,                   &
                      "start3",istart3,"end3",iend3 
            end if
#endif

            ibase = 1
            ivol = 0
            
            !2D xy domain    
            if (nvxgbl > 1 .and. nvygbl > 1) then
                dmda_flow%xlmin=lmin(1)
                dmda_flow%xlmax=lmax(1)
                dmda_flow%ylmin=lmin(2)
                dmda_flow%ylmax=lmax(2)
                
                dmda_flow%xlmingbl=gmin(1)
                dmda_flow%xlmaxgbl=gmax(1)
                dmda_flow%ylmingbl=gmin(2)
                dmda_flow%ylmaxgbl=gmax(2)
#ifdef DEBUG
                if(info_debug > 0) then
                    do ivy = nvyls,nvyle
                        do ivx = nvxls,nvxle
                          ivol = (ivy-nvygls)*nvxgl + ivx-nvxgls+1 
                          write(idbg,                                  &
                               '(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))') &
                               "ivol_l", ivol,                         &
                               "ivol_g", node_idx_lg2g(ivol),          &
                               "xg", coords2d(1-ibase,ivx-ibase,       &
                                              ivy-ibase),              &
                               "yg", coords2d(2-ibase,ivx-ibase,       &
                                              ivy-ibase),              &
                               "zg", r0 
                        end do
                    end do
                end if
#endif
            !2D yz domain    
            else if (nvygbl > 1 .and. nvzgbl > 1) then  
                dmda_flow%ylmin=lmin(1)
                dmda_flow%ylmax=lmax(1)
                dmda_flow%zlmin=lmin(2)
                dmda_flow%zlmax=lmax(2)
                
                dmda_flow%ylmingbl=gmin(1)
                dmda_flow%ylmaxgbl=gmax(1)
                dmda_flow%zlmingbl=gmin(2)
                dmda_flow%zlmaxgbl=gmax(2)
#ifdef DEBUG
                if(info_debug > 0) then
                    do ivz = nvzls, nvzle
                        do ivy = nvyls,nvyle   
                          ivol = (ivz-nvzgls)*nvygl + ivy-nvygls+1    
                          write(idbg,                                  &
                               '(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))') &
                               "ivol_l", ivol,                         &
                               "ivol_g", node_idx_lg2g(ivol),          &
                               "xg", r0,                               &
                               "yg", coords2d(1-ibase,ivy-ibase,       &
                                              ivz-ibase),              &
                               "zg", coords2d(2-ibase,ivy-ibase,       &
                                              ivz-ibase)   
                        end do
                    end do
                end if
#endif
            !2D xz domain    
            else if (nvzgbl > 1 .and. nvxgbl > 1) then
                dmda_flow%xlmin=lmin(1)
                dmda_flow%xlmax=lmax(1)
                dmda_flow%zlmin=lmin(2)
                dmda_flow%zlmax=lmax(2)
                
                dmda_flow%xlmingbl=gmin(1)
                dmda_flow%xlmaxgbl=gmax(1)
                dmda_flow%zlmingbl=gmin(2)
                dmda_flow%zlmaxgbl=gmax(2)
#ifdef DEBUG
                if(info_debug > 0) then
                    do ivz = nvzls, nvzle
                        do ivx = nvxls,nvxle 
                          ivol = (ivz-nvzgls)*nvxgl + ivx-nvxgls+1   
                          write(idbg,                                  &
                               '(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))') &
                               "ivol_l", ivol,                         &
                               "ivol_g", node_idx_lg2g(ivol),          &
                               "xg", coords2d(1-ibase,ivx-ibase,       &
                                              ivz-ibase),              &
                               "yg", r0,                               &
                               "zg", coords2d(2-ibase,ivx-ibase,       &
                                              ivz-ibase)
                        end do
                    end do
                end if
#endif
            end if
            
            call DMDAVecRestoreArrayF90(cda,gc,coords2d,ierr)
            CHKERRQ(ierr)
            
        case (3) 
#ifdef DEBUG
            if(info_debug > 0) then
                write(idbg,'(2(/a/,6(a,1x,1pe15.6e3,1x)))')            &
                      "local bounding box 3 dim",                      &
                      "lmin",lmin(1),"lmax",lmax(1),                   &
                      "lmin",lmin(2),"lmax",lmax(2),                   &
                      "lmin",lmin(3),"lmax",lmax(3),                   &
                      "global bounding box 3 dim",                     &
                      "gmin",gmin(1),"gmax",gmax(1),                   &
                      "gmin",gmin(2),"gmax",gmax(2),                   &
                      "gmin",gmin(3),"gmax",gmax(3)  
            end if
#endif
            call DMDAVecGetArrayF90(cda,gc,coords3d, ierr)
            CHKERRQ(ierr)
        
            dmda_flow%xlmin=lmin(1)
            dmda_flow%xlmax=lmax(1)
            dmda_flow%ylmin=lmin(2)
            dmda_flow%ylmax=lmax(2)
            dmda_flow%zlmin=lmin(3)
            dmda_flow%zlmax=lmax(3)
            
            dmda_flow%xlmingbl=gmin(1)
            dmda_flow%xlmaxgbl=gmax(1)
            dmda_flow%ylmingbl=gmin(2)
            dmda_flow%ylmaxgbl=gmax(2)
            dmda_flow%zlmingbl=gmin(3)
            dmda_flow%zlmaxgbl=gmax(3)
            
            !Please note the index is 0-based
#ifdef DEBUG
            if(info_debug > 0) then
                istart1 = lbound(coords3d,1)
                iend1 = ubound(coords3d,1)  
            
                istart2 = lbound(coords3d,2)
                iend2 = ubound(coords3d,2) 
            
                istart3 = lbound(coords3d,3)
                iend3 = ubound(coords3d,3) 
            
                istart4 = lbound(coords3d,4)
                iend4 = ubound(coords3d,4)
        
                write(idbg,'(a/,8(a,1x,i10,1x))')                      &
                      "get the coordinates 3 dim",                     &
                      "start1",istart1,"end1",iend1,                   &
                      "start2",istart2,"end2",iend2,                   &
                      "start3",istart3,"end3",iend3,                   &
                      "start4",istart4,"end4",iend4
            end if
#endif
            ibase = 1
            ivol = 0
#ifdef DEBUG
            if(info_debug > 0) then
                do ivz = nvzls, nvzle
                    do ivy = nvyls,nvyle
                        do ivx = nvxls,nvxle   
                          ivol = (ivz-nvzgls)*nvygl*nvxgl +            &
                                 (ivy-nvygls)*nvxgl + ivx-nvxgls+1  
                          write(idbg,                                  &
                                '(2(a,1x,i6,1x),3(a,1x,1pe15.6e3,1x))')&
                                "ivol_l", ivol,                        &
                                "ivol_g", node_idx_lg2g(ivol),         &
                                "xg", coords3d(1-ibase,ivx-ibase,      &
                                         ivy-ibase,ivz-ibase),         &
                                "yg", coords3d(2-ibase,ivx-ibase,      &
                                         ivy-ibase,ivz-ibase),         &
                                "zg", coords3d(3-ibase,ivx-ibase,      &
                                         ivy-ibase,ivz-ibase)
                        end do
                    end do
                end do
            end if
#endif
            call DMDAVecRestoreArrayF90(cda,gc,coords3d,ierr)
            CHKERRQ(ierr)
          
        end select

    end subroutine solver_dd_coordinates_get
    
    !>
    !> mykspconverged_flow
    !> This is a user-defined routine for testing
    !> convergence of the KSP iterative solvers.
    !>  Input Parameters:
    !>    ksp   - iterative context
    !>    n     - iteration number
    !>    rnorm - 2-norm (preconditioned) residual value (may be estimated)
    !>    dummy - optional user-defined monitor context (unused here)
    subroutine mykspconverged_flow(ksp_flow,n,rnorm,flag,dummy,ierr)
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscksp.h>
        use petscsys
        use petscksp
#endif
        use gen, only : rank
    
        implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscksp.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscksp.h>
#endif
        
        KSP                :: ksp_flow
        PetscInt           :: n,dummy
        KSPConvergedReason :: flag
        PetscReal          :: rnorm
        PetscErrorCode     :: ierr

        Vec                :: x_flow
        PetscReal          :: rnorm_relative
        PetscReal          :: neg_one
       
        neg_one = -1.0d0
        
        if(n == 0) then
            rnorm_init_flow = rnorm
        end if
        
        rnorm_relative = rnorm / rnorm_init_flow
       
        !Get true residual norm
        !This operation is expensive, use -ksp_norm_type unpreconditioned instead
        if(n > 0) then
            call KSPBuildSolution(ksp_flow,                            &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                                  PETSC_NULL_VEC,                      &
#else
                                  PETSC_NULL_OBJECT,                   &
#endif
                                  x_flow, ierr)
            CHKERRQ(ierr)
            call KSPGetRhs(ksp_flow, b_flow, ierr)
            CHKERRQ(ierr)
            call KSPGetOperators(ksp_flow,a_flow,                      &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                                 PETSC_NULL_MAT,                       &
#else
                                 PETSC_NULL_OBJECT,PETSC_NULL_OBJECT,  &
#endif
                                 ierr)
            CHKERRQ(ierr)
            
            call calculate_norm(a_flow, x_flow, b_flow, r_flow,        &
                                rnorm_flow)
            
        else
            rnorm_flow = 1.0d30
        end if

        if (n > 0 .and. (rnorm_flow < abstol_flow .or.                 &
            rnorm_relative < rtol_flow)) then
          flag = 1
#ifdef DEBUG
          if(rank == 0) then
              if(rnorm_flow < abstol_flow .and.                        &
                      rnorm_relative < rtol_flow) then
                  write(*,'(2(1x, a, 1x, 1pe15.6e3))')                 &
                        "KSP converged due to abstol", rnorm_flow,     &
                        "and rstol", rnorm_relative
              else if (rnorm_flow < abstol_flow) then
                  write(*,'(1x, a, 1x, 1pe15.6e3)')                    &
                        "KSP converged due to abstol", rnorm_flow
              else if (rnorm_relative < rtol_flow) then
                  write(*,'(1x, a, 1x, 1pe15.6e3)')                    &
                        "KSP converged due to rstol", rnorm_relative
              end if
          end if
#endif
        else
          flag = 0
        endif
        
        ierr = 0

    end subroutine mykspconverged_flow
    
    !>
    !> mykspconverged_heat
    !> This is a user-defined routine for testing
    !> convergence of the KSP iterative solvers.
    !>  Input Parameters:
    !>    ksp   - iterative context
    !>    n     - iteration number
    !>    rnorm - 2-norm (preconditioned) residual value (may be estimated)
    !>    dummy - optional user-defined monitor context (unused here)
    subroutine mykspconverged_heat(ksp_heat,n,rnorm,flag,dummy,ierr)
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscksp.h>
        use petscsys
        use petscksp
#endif
        use gen, only : rank
    
        implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscksp.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscksp.h>
#endif
        
        KSP                :: ksp_heat
        PetscInt           :: n,dummy
        KSPConvergedReason :: flag
        PetscReal          :: rnorm
        PetscErrorCode     :: ierr

        Vec                :: x_heat
        PetscReal          :: rnorm_relative
        PetscReal          :: neg_one
       
        neg_one = -1.0d0
        
        if(n == 0) then
            rnorm_init_heat = rnorm
        end if
        
        rnorm_relative = rnorm / rnorm_init_heat
       
        !Get true residual norm
        !This operation is expensive, use -ksp_norm_type unpreconditioned instead
        if(n > 0) then
            call KSPBuildSolution(ksp_heat,                            &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                                  PETSC_NULL_VEC,                      &
#else
                                  PETSC_NULL_OBJECT,                   &
#endif
                                  x_heat, ierr)
            CHKERRQ(ierr)
            call KSPGetRhs(ksp_heat, b_heat, ierr)
            CHKERRQ(ierr)
            call KSPGetOperators(ksp_heat,a_heat,                      &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                                 PETSC_NULL_MAT,                       &
#else
                                 PETSC_NULL_OBJECT,PETSC_NULL_OBJECT,  &
#endif
                                 ierr)
            CHKERRQ(ierr)
            
            call calculate_norm(a_heat, x_heat, b_heat, r_heat,        &
                                rnorm_heat)
            
        else
            rnorm_heat = 1.0d30
        end if

        if (n > 0 .and. (rnorm_heat < abstol_heat .or.                 &
            rnorm_relative < rtol_heat)) then
          flag = 1
#ifdef DEBUG
          if(rank == 0) then
              if(rnorm_heat < abstol_heat .and.                        &
                      rnorm_relative < rtol_heat) then
                  write(*,'(2(1x, a, 1x, 1pe15.6e3))')                 &
                        "KSP converged due to abstol", rnorm_heat,     &
                        "and rstol", rnorm_relative
              else if (rnorm_heat < abstol_heat) then
                  write(*,'(1x, a, 1x, 1pe15.6e3)')                    &
                        "KSP converged due to abstol", rnorm_heat
              else if (rnorm_relative < rtol_heat) then
                  write(*,'(1x, a, 1x, 1pe15.6e3)')                    &
                        "KSP converged due to rstol", rnorm_relative
              end if
          end if
#endif
        else
          flag = 0
        endif
        
        ierr = 0

    end subroutine mykspconverged_heat    
    !>
    !> mykspconverged_react
    !> This is a user-defined routine for testing
    !> convergence of the KSP iterative solvers.
    !>  Input Parameters:
    !>    ksp   - iterative context
    !>    n     - iteration number
    !>    rnorm - 2-norm (preconditioned) residual value (may be estimated)
    !>    dummy - optional user-defined monitor context (unused here)
    subroutine mykspconverged_react(ksp_react,n,rnorm,flag,dummy,ierr)

#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
        use petscsys
#endif
        use gen, only : rank
    
        implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscksp.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscksp.h>
#endif
        
        KSP                :: ksp_react
        Vec                :: x_react
        PetscErrorCode     :: ierr
        PetscInt           :: n,dummy
        KSPConvergedReason :: flag
        PetscReal          :: rnorm
        PetscReal          :: rnorm_relative
        PetscReal          :: neg_one
       
        neg_one = -1.0d0
        
        if(n == 0) then
            rnorm_init_react = rnorm
        end if
        
        rnorm_relative = rnorm / rnorm_init_react
       
        !Get true residual norm
        !This operation is expensive, use -ksp_norm_type unpreconditioned instead
        if(n > 0) then
            call KSPBuildSolution(ksp_react,                           &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                                  PETSC_NULL_VEC,                      &
#else
                                  PETSC_NULL_OBJECT,                   &
#endif
                                  x_react, ierr)
            CHKERRQ(ierr)

            call KSPGetRhs(ksp_react, b_react, ierr)
            CHKERRQ(ierr)

            call KSPGetOperators(ksp_react,a_react,                    &
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
                                 PETSC_NULL_MAT,                       &
#else
                                 PETSC_NULL_OBJECT,PETSC_NULL_OBJECT,  &
#endif
                                 ierr)
            CHKERRQ(ierr)
            
            call calculate_norm(a_react, x_react, b_react, r_react,    &
                                rnorm_react)
            
        else
            rnorm_react = 1.0d30
        end if

        if (n > 0 .and. (rnorm_react < abstol_react .or.               &
            rnorm_relative < rtol_react)) then
          flag = 1
#ifdef DEBUG
          if(rank == 0) then
              if(rnorm_react < abstol_react .and.                      &
                      rnorm_relative < rtol_react) then
                  write(*,'(2(1x, a, 1x, 1pe15.6e3))')                 &
                        "KSP converged due to abstol", rnorm_react,    &
                        "and rstol", rnorm_relative
              else if (rnorm_react < abstol_react) then
                  write(*,'(1x, a, 1x, 1pe15.6e3)')                    &
                        "KSP converged due to abstol", rnorm_react
              else if (rnorm_relative < rtol_react) then
                  write(*,'(1x, a, 1x, 1pe15.6e3)')                    &
                        "KSP converged due to rstol", rnorm_relative
              end if
          end if
#endif
        else
          flag = 0
        endif
        
        ierr = 0

    end subroutine mykspconverged_react
    

    !>
    !> calculate norm for the current solution
    !>
    subroutine calculate_norm(a, x, b, u, rnorm)

#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
      use petscsys
#endif

      implicit none

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif

      ! passing variables
      Mat                :: a
      Vec                :: x, b, u
      PetscReal          :: rnorm
      PetscReal          :: neg_one

      ! local variables
      PetscErrorCode     :: ierr

      neg_one = -1.0d0

      rnorm = 1.0d30

      call MatMult(a, x, u, ierr)
      CHKERRQ(ierr)

      call VecAXPY(u, neg_one, b, ierr)
      CHKERRQ(ierr)

      call VecNorm(u,NORM_2,rnorm,ierr)
      CHKERRQ(ierr)

    end subroutine calculate_norm

    
    !> Release solver space
    subroutine solver_dd_release_heat

#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
        use petscsys
#endif


        implicit none
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif
        
        PetscErrorCode :: ierr
        
        call MatDestroy(a_heat,ierr)
        CHKERRQ(ierr)

        if (b_reuse_preconditioner_heat) then
          call MatDestroy(a_heat_j,ierr)
          CHKERRQ(ierr)
        end if

        call VecDestroy(x_heat_loc,ierr)
        CHKERRQ(ierr)
        call VecDestroy(x_heat,ierr)
        CHKERRQ(ierr)
        call VecDestroy(r_heat_loc,ierr)
        CHKERRQ(ierr)
        call VecDestroy(r_heat,ierr)
        CHKERRQ(ierr)
        call VecDestroy(b_heat_loc,ierr)
        CHKERRQ(ierr)
        call VecDestroy(b_heat,ierr)
        CHKERRQ(ierr)

        if (trim(strKSPType_heat) == "kspsuperlu" .or. &
            trim(strKSPType_heat) == "superlu"    .or. &
            trim(strKSPType_heat) == "kspmumps"   .or. &
            trim(strKSPType_heat) == "mpmps") then
          call MatDestroy(a_heat_fac,ierr)
          CHKERRQ(ierr)
        else
          call KSPDestroy(ksp_heat,ierr)
          CHKERRQ(ierr)
        end if

   
    end subroutine solver_dd_release_heat    

    !> Release solver space
    subroutine solver_dd_release_react

#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
        use petscsys
#endif

        use gen, only: reactive_transport
    
        implicit none

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif
        PetscErrorCode :: ierr

        if(reactive_transport) then
            call MatDestroy(a_react,ierr)
            CHKERRQ(ierr)

            if (b_reuse_preconditioner_react) then
              call MatDestroy(a_react_j,ierr)
              CHKERRQ(ierr)
            end if

            call VecDestroy(x_react_loc,ierr)
            CHKERRQ(ierr)
            call VecDestroy(x_react,ierr)
            CHKERRQ(ierr)
            call VecDestroy(r_react_loc,ierr)
            CHKERRQ(ierr)
            call VecDestroy(r_react,ierr)
            CHKERRQ(ierr)
            call VecDestroy(b_react_loc,ierr)
            CHKERRQ(ierr)
            call VecDestroy(b_react,ierr)
            CHKERRQ(ierr)

            if (trim(strKSPType_react) == "kspsuperlu" .or. &
                trim(strKSPType_react) == "superlu"    .or. &
                trim(strKSPType_react) == "kspmumps"   .or. &
                trim(strKSPType_react) == "mpmps") then
              call MatDestroy(a_react_fac,ierr)
              CHKERRQ(ierr)
            else
              call KSPDestroy(ksp_react,ierr)
              CHKERRQ(ierr)
            end if

        end if
   
    end subroutine solver_dd_release_react

end module solver_dd
    
#endif
