!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 826 $
!> $Author: dsu $
!> $Date: 2022-03-24 10:10:16 -0700 (Thu, 24 Mar 2022) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/solver/solver_snes_common.F90 $
!---------------------------------------------------------------------
!********************************************************************!

#ifdef PETSC

!> module: solver_snes_common
!>
!> written by: Danyang Su
!>
!> module description:
!>
!> Module of nonlinear function  
!>
!> Note:
!> See http://www.mcs.anl.gov/petsc/ for detail


module solver_snes_common
#ifdef PETSC
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscksp.h>
#ifdef PETSC_USE_LOG
#include <petsc/finclude/petsclog.h>
#endif
    use petscsys
    use petscdmda
    use petscksp
#endif
#endif

    implicit none

#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsysdef.h>
#include <petsc/finclude/petscvecdef.h>
#include <petsc/finclude/petscdmdef.h>
#include <petsc/finclude/petscsys.h>
#include <petsc/finclude/petscvec.h>
#include <petsc/finclude/petscdmda.h>
#include <petsc/finclude/petscis.h>
#include <petsc/finclude/petscmat.h>
#include <petsc/finclude/petscksp.h>
#include <petsc/finclude/petscpc.h>
#include <petsc/finclude/petscvec.h90>
#include <petsc/finclude/petscdmda.h90>
#ifdef PETSC_USE_LOG
#include <petsc/finclude/petsclog.h>
#endif
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsysdef.h>
#include <finclude/petscvecdef.h>
#include <finclude/petscdmdef.h>
#include <finclude/petscsys.h>
#include <finclude/petscvec.h>
#include <finclude/petscdmda.h>
#include <finclude/petscis.h>
#include <finclude/petscmat.h>
#include <finclude/petscksp.h>
#include <finclude/petscpc.h>
#include <finclude/petscvec.h90>
#include <finclude/petscdmda.h90>
#ifdef PETSC_USE_LOG
#include <finclude/petsclog.h>
#endif
#endif

       type userctx
        !> DMDA
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 11)
        type(tDM) :: da
#else
        DM :: da
#endif
        !> degree of freedom
        PetscInt :: dof
        !> stencil width
        PetscInt :: swidth
        !> dimension of the distributed array
        PetscInt :: dim
        !> grid information
        PetscInt :: xs,xe,xl,gxs,gxe,gxl
        PetscInt :: ys,ye,yl,gys,gye,gyl
        PetscInt :: zs,ze,zl,gzs,gze,gzl
        !> global Vector ranger, 1-based  
        PetscInt :: range_start, range_end
        !> global dimension in each direction of the array
        PetscInt :: nvxgbl,nvygbl,nvzgbl
        !> corresponding number of procs in each dimension
        PetscInt :: npx, npy, npz  
        !> local bounding box coordinates in each dimension
        PetscReal :: xlmin, xlmax, ylmin, ylmax, zlmin, zlmax
        !> global bounding box coordinates in each dimension
        PetscReal :: xlmingbl, xlmaxgbl, ylmingbl, ylmaxgbl, zlmingbl, zlmaxgbl
       
        !Other parameters, to be removed if not necessary
        double precision lambda  
      end type userctx   
      
      !> Flag indicating that intermediate mesh entities (faces, edges) should be created automatically.
      !> Importnant note: for prism mesh, please set interpolate to true, and change index after distribution
      !> This parameter is set to true by default after April 8, 2020
      !> Update on 2021-04-18: Hexa mesh cannot use interpolate = true feature, otherwise, 
      !>                       domain decomposition is not efficient.
      PetscBool :: interpolate

      !> General variables
      PetscInt :: stencil_width

      !> PETSc partitioner type
      character(16) :: partitioner_type

      !PetscErrorCode :: ierr
    
      !> Variables of flow problem
      !> DMDA 
      type(userctx), target :: dmda_flow 
      !> KSP
      KSP, target :: ksp_flow

      !> Solution
      Vec, target :: x_flow
      Vec, target :: x_flow_loc
      !> Residual
      Vec, target :: r_flow
      Vec, target :: r_flow_loc
      !> Right hand side
      Vec, target :: b_flow
      Vec, target :: b_flow_loc
      !> Jacobian matrix
      Mat, target :: a_flow
      !> Matrix operator J    
      Mat, target :: a_flow_j
      !> Factored matrix
      Mat, target :: a_flow_fac
      
      !> Parameters for PC flow 
      PC, target :: pc_flow

      !> incomplete factorization level for PC flow
      PetscInt :: pc_factor_level_flow

     
      !> @strKSPType PETSc KSP solver type, e.g, "KSPGMRES", "KSPBCGS"
      character(16) :: strKSPType_flow
      
      !> @strKSPNormType PETSc KSP Norm type, e.g., "KSP_NORM_NONE", "KSP_NORM_PRECONDITIONED"
      !>                                            "KSP_NORM_UNPRECONDITIONED", "KSP_NORM_NATURAL"
      character(32) :: strKSPNormType_flow
      
      !> @strPCType_flow PETSc preconditioner type in flow, e.g., "pcbjacobi", "pcilu"
      character(32) :: strPCType_flow
      
      !> @strKSPConvergenceType_flow PETSc convergence criteria type in flow, e.g., "kspdefault", "kspuserdefined"
      character(32) :: strKSPConvergenceType_flow

      !> @param PC factor shift type
      character(32) :: pc_factor_shift_flow         !default none
      
      !> @b_mykspconverged_flow
      PetscBool, target :: b_mykspconverged_flow
      
      !> @b_recal_norm_flow
      PetscBool, target :: b_recal_norm_flow
      
      !> @b_check_norm_flow
      PetscBool, target :: b_check_norm_flow
      
      !> @param rtol the relative convergence tolerance (relative decrease in the residual norm) 
      PetscReal :: rtol_flow         !default 1.0E-5
      !> @param abstol the absolute convergence tolerance (absolute size of the residual norm) 
      PetscReal :: abstol_flow       !default 1.0E-30
      !> @param dtol  the divergence tolerance (amount residual can increase before KSPDefaultConverged() concludes that the method is diverging) 
      PetscReal :: dtol_flow         !default 1.0E5
      !> @param maxits maximum number of iterations to use    
      PetscInt  :: maxits_flow       !default 1.0E5
      !> @param residual norm, can be preconditioned residual norm or true residual norm
      PetscReal, target :: rnorm_flow
      !> @param initial residual norm, can be preconditioned residual norm or true residual norm
      PetscReal, target :: rnorm_init_flow
      
      !> @param b_use_petsc_default_flow
      PetscBool :: b_use_petsc_default_flow
      
      !> @param b_initial_guess_nonzero_flow
      PetscBool :: b_initial_guess_nonzero_flow

      !> @param b_form_initial_guess_flow
      PetscBool :: b_form_initial_guess_flow

      !> @param b_reuse_preconditioner_flow
      PetscBool :: b_reuse_preconditioner_flow
      
      !> @param b_set_preconditioner_flow
      PetscBool, target :: b_set_preconditioner_flow

#ifdef PETSC_USE_LOG
      !> @param profiling of initial guess for flow
      PetscLogEvent :: log_x_flow
      !> @param profiling of rhs for flow
      PetscLogEvent :: log_b_flow
      !> @param profiling of initial rhs norm2 for flow
      PetscLogEvent :: log_bnorm_flow
      !> @param profiling of jacobi matrix for flow
      PetscLogEvent :: log_a_flow
      !> @param profiling of solver for flow
      PetscLogEvent :: log_ksp_flow
      !> @param profiling of total ksp solver for flow
      PetscLogEvent :: log_tot_flow
#endif

      !> Variables of decoupled heat problem
      !> DMDA 
      type(userctx), target :: dmda_heat 
      !> KSP
      KSP, target :: ksp_heat

      !> Solution
      Vec, target :: x_heat
      Vec, target :: x_heat_loc
      !> Residual
      Vec, target :: r_heat
      Vec, target :: r_heat_loc
      !> Right hand side
      Vec, target :: b_heat
      Vec, target :: b_heat_loc
      !> Jacobian matrix
      Mat, target :: a_heat
      !> Matrix operator J    
      Mat, target :: a_heat_j
      !> Factored matrix
      Mat, target :: a_heat_fac
      
      !> Parameters for PC heat 
      PC, target :: pc_heat

      !> incomplete factorization level for PC heat
      PetscInt :: pc_factor_level_heat

     
      !> @strKSPType PETSc KSP solver type, e.g, "KSPGMRES", "KSPBCGS"
      character(16) :: strKSPType_heat
      
      !> @strKSPNormType PETSc KSP Norm type, e.g., "KSP_NORM_NONE", "KSP_NORM_PRECONDITIONED"
      !>                                            "KSP_NORM_UNPRECONDITIONED", "KSP_NORM_NATURAL"
      character(32) :: strKSPNormType_heat
      
      !> @strPCType_heat PETSc preconditioner type in heat, e.g., "pcbjacobi", "pcilu"
      character(32) :: strPCType_heat
      
      !> @strKSPConvergenceType_heat PETSc convergence criteria type in heat, e.g., "kspdefault", "kspuserdefined"
      character(32) :: strKSPConvergenceType_heat

      !> @param PC factor shift type
      character(32) :: pc_factor_shift_heat         !default none
      
      !> @b_mykspconverged_heat
      PetscBool, target :: b_mykspconverged_heat
      
      !> @b_recal_norm_heat
      PetscBool, target :: b_recal_norm_heat
      
      !> @b_check_norm_heat
      PetscBool, target :: b_check_norm_heat
      
      !> @param rtol the relative convergence tolerance (relative decrease in the residual norm) 
      PetscReal :: rtol_heat         !default 1.0E-5
      !> @param abstol the absolute convergence tolerance (absolute size of the residual norm) 
      PetscReal :: abstol_heat       !default 1.0E-30
      !> @param dtol  the divergence tolerance (amount residual can increase before KSPDefaultConverged() concludes that the method is diverging) 
      PetscReal :: dtol_heat         !default 1.0E5
      !> @param maxits maximum number of iterations to use    
      PetscInt  :: maxits_heat       !default 1.0E5
      !> @param residual norm, can be preconditioned residual norm or true residual norm
      PetscReal, target :: rnorm_heat
      !> @param initial residual norm, can be preconditioned residual norm or true residual norm
      PetscReal, target :: rnorm_init_heat
      
      !> @param b_use_petsc_default_heat
      PetscBool :: b_use_petsc_default_heat
      
      !> @param b_initial_guess_nonzero_heat
      PetscBool :: b_initial_guess_nonzero_heat

      !> @param b_form_initial_guess_heat
      PetscBool :: b_form_initial_guess_heat

      !> @param b_reuse_preconditioner_heat
      PetscBool :: b_reuse_preconditioner_heat
      
      !> @param b_set_preconditioner_heat
      PetscBool, target :: b_set_preconditioner_heat

#ifdef PETSC_USE_LOG
      !> @param profiling of initial guess for heat
      PetscLogEvent :: log_x_heat
      !> @param profiling of rhs for heat
      PetscLogEvent :: log_b_heat
      !> @param profiling of initial rhs norm2 for heat
      PetscLogEvent :: log_bnorm_heat
      !> @param profiling of jacobi matrix for heat
      PetscLogEvent :: log_a_heat
      !> @param profiling of solver for heat
      PetscLogEvent :: log_ksp_heat
      !> @param profiling of total ksp solver for heat
      PetscLogEvent :: log_tot_heat
#endif

      !> Varialbes of reactive transport problem
      !> DMDA  
      type(userctx) :: dmda_react
      !> KSP
      KSP, target :: ksp_react

      !> Solution
      Vec, target :: x_react
      Vec, target :: x_react_loc
      !> Residual
      Vec, target :: r_react
      Vec, target :: r_react_loc
      !> Right hand side
      Vec, target :: b_react
      Vec, target :: b_react_loc
      !> Jacobian matrix
      Mat, target :: a_react
      !> Matrix operator J    
      Mat, target :: a_react_j
      !> Factored matrix
      Mat, target :: a_react_fac
      
      !> Parameters for PC reactive transport
      PC, target :: pc_react
      
      !> incomplete factorization level for PC reactive transport
      PetscInt :: pc_factor_level_react

      !> @strKSPType PETSc KSP solver type, e.g, "KSPGMRES", "KSPBCGS"
      character(16) :: strKSPType_react
      
      !> @strKSPNormType PETSc KSP Norm type, e.g., "KSP_NORM_NONE", "KSP_NORM_PRECONDITIONED"
      !>                                            "KSP_NORM_UNPRECONDITIONED", "KSP_NORM_NATURAL"
      character(32) :: strKSPNormType_react
      
      !> @strPCType_react PETSc preconditioner type in reactive transport, e.g., "pcbjacobi", "pcilu"
      character(32) :: strPCType_react
      
      !> @strKSPConvergenceType_react PETSc convergence criteria type in reactive transport, e.g., "kspdefault", "kspuserdefined"
      character(32) :: strKSPConvergenceType_react

      !> @param PC factor shift type
      character(32) :: pc_factor_shift_react         !default none
      
      !> @b_mykspconverged_react
      PetscBool, target :: b_mykspconverged_react
      
      !> @b_recal_norm_react
      PetscBool, target :: b_recal_norm_react
      
      !> @b_check_norm_react
      PetscBool, target :: b_check_norm_react
      
      !> @param rtol the relative convergence tolerance (relative decrease in the residual norm) 
      PetscReal :: rtol_react         !default 1.0E-5
      !> @param abstol the absolute convergence tolerance (absolute size of the residual norm) 
      PetscReal :: abstol_react       !default 1.0E-30
      !> @param dtol  the divergence tolerance (amount residual can increase before KSPDefaultConverged() concludes that the method is diverging) 
      PetscReal :: dtol_react         !default 1.0E5
      !> @param maxits maximum number of iterations to use    
      PetscInt :: maxits_react       !default 1.0E5
      !> @param residual norm, can be preconditioned residual norm or true residual norm
      PetscReal, target :: rnorm_react
      !> @param initial residual norm, can be preconditioned residual norm or true residual norm
      PetscReal, target :: rnorm_init_react

      !> @param b_use_petsc_default_react
      PetscBool :: b_use_petsc_default_react

      !> @param b_initial_guess_nonzero_react
      PetscBool :: b_initial_guess_nonzero_react
      
      !> @param b_form_initial_guess_react
      PetscBool :: b_form_initial_guess_react

      !> @param b_reuse_preconditioner_react
      PetscBool :: b_reuse_preconditioner_react
      
      !> @param b_set_preconditioner_react
      PetscBool, target :: b_set_preconditioner_react

#ifdef PETSC_USE_LOG
      !> @param profiling of initial guess for reactive transport
      PetscLogEvent :: log_x_react
      !> @param profiling of rhs for reactive transport
      PetscLogEvent :: log_b_react
      !> @param profiling of initial rhs norm2 for flow
      PetscLogEvent :: log_bnorm_react
      !> @param profiling of jacobi matrix for reactive transport
      PetscLogEvent :: log_a_react
      !> @param profiling of solver for reactive transport
      PetscLogEvent :: log_ksp_react
      !> @param profiling of total ksp solver for reactive transport
      PetscLogEvent :: log_tot_react
#endif

end module solver_snes_common
    
#endif
