!*****Revision Informations Automatically Generated by VisualSVN*****!
!---------------------------------------------------------------------
!> $ID:$
!> $Revision: 850 $
!> $Author: dsu $
!> $Date: 2023-01-27 08:58:23 -0800 (Fri, 27 Jan 2023) $
!> $URL: https://biot.eos.ubc.ca/svn/min3p_thcm/branches/dsu_new_add_2023Jan/src/min3p/sptldisc.F90 $
!---------------------------------------------------------------------
!********************************************************************!

!c ----------------------------------------------------------------------
!c subroutine sptldisc
!c -------------------
!c
!c read spatial discretization parameters,
!c discretize solution domain and compute nodal volumes
!c
!c written by:      Uli Mayer - May 11, 96
!c
!c last modified:   Uli Mayer - November 25, 96
!c
!c                  Danyang Su - Sept. 10, 2018
!c                  Unstructured grid and HPC capabilities
!c
!c             Celine Blitz Frayret (CBF) for Frederic Gerard, December 14, 2018 :
!c            addition of toparea
!c
!c definition of variables:
!c
!c I --> on input   * arbitrary  - initialized  + entries expected
!c O --> on output  * arbitrary  - unaltered    + altered
!c 
!c                                                                    I O
!c passed:   -
!c
!c common:   
!c parm.f:   -
!c
!c gen.f:    real*8:
!c           -------
!c           xmax(nxx)          = max. coordinate of discretization   * +
!c                                interval in x-direction
!c           xmin(nxx)          = min. coordinate of discretization   * +
!c                                interval in x-direction
!c           ymax(nyy)          = max. coordinate of discretization   * +
!c                                interval in y-direction
!c           ymin(nyy)          = min. extent of solution domain      * +
!c                                in y-direction
!c           zmax(nzz)          = max. extent of solution domain      * +
!c                                in z-direction
!c           zmin(nzz)          = min. extent of solution domain      * +
!c                                in z-direction
!c           cvol(nn)           = nodal volumes                       * +
!c           dimcv(3,nn)        = dimension of control volumes        * +
!c           xg(nn)             = spatial coordinates in x-direction  * +
!c           yg(nn)             = spatial coordinates in y-direction  * +
!c           zg(nn)             = spatial coordinates in z-direction  * +
!c           tkel(nn)           = nodal temperatures in Kelvin        * +
!c           delx(nvx)          = spatial increment in x-direction    * +
!c           dely(nvy)          = spatial increment in y-direction    * +
!c           delz(nvz)          = spatial increment in z-direction    * +
!c
!c           integer*4:
!c           ----------
!c           idat               = unit number, run specific inputs    + -
!c                                             file
!c           igen               = unit number, generic output file    + -
!c           ilog               = unit number, log file               + -
!c           itmp               = unit number, temporary storage      + -
!c           nxx                = number of intervals in x-direction  * +
!c           nyy                = number of intervals in y-direction  * +
!c           nzz                = number of intervals in z-direction  * +
!c           nvix(nxx)          = number of control volumes in        * +
!c                                x-direction (intervals)
!c           nviy(nyy)          = number of control volumes in        * +
!c                                y-direction (intervals)
!c           nviz(nzz)          = number of control volumes in        * +
!c                                z-direction (intervals)
!c           nvx                = number of control volumes in        * +
!c                                x-direction
!c           nvy                = number of control volumes in        * +
!c                                y-direction
!c           nvz                = number of control volumes in        * +
!c                                z-direction
!c           nn                 = total number of control volumes     * +
!c
!c           logical:
!c           --------
!c           half_cells         = .true.  -> half cells on boundary   * +
!c
!c           character:
!c           ----------
!c           section_header     = section header                      * *
!c
!c local:    integer*4:
!c           ----------
!c           ierr               = 0 -> memory allocation successful
!c           ixx                = counter (intervals in x-direction)
!c           iyy                = counter (intervals in y-direction)
!c           izz                = counter (intervals in z-direction)
!c           l_string           = length of text string
!c
!c           real*8:
!c           -------
!c           xtot               = total dimension in x-direction  CBF 
!c           ytot               = total dimension in y-direction  CBF 
!c
!c           logical:
!c           --------
!c           found_section      = .true.  -> section header was
!c                                           found in input file
!c
!c external: checkerr  = check for error during memory allocation 
!c           cvolume   = compute volumes on a nodal basis
!c           readbloc  = read section of input file and write to
!c                       temporary file
!c           xyzcoord  = compute spatial increments for x,y and 
!c                       z-direction and calculate xyz-coordinates 
!c                       of control volume centroids
!c---------------------------------------------------------------------- 
!c WARNING
!c---------------------------------------------------------------------- 
!c All allocated variables are initialized to some value. This depends  
!c of its type:
!c 
!c real*8      => 0.0d0
!c integer     => 0 
!c character   => ' '    
!c logical     => .false. 
!c 
!c Sergio Andres Bea Jofre (2009)
!c
!c ----------------------------------------------------------------------
 
      subroutine sptldisc

#ifdef PETSC
#include <petscversion.h>
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 8)
#include <petsc/finclude/petscsys.h>
      use petscsys
#endif
#endif

      use parm
      use gen
      use file_unit, only : lun_get, lun_free
#ifdef PETSC
      use solver_dd, only : solver_dd_DMDACreate_flow_heat
      use solver_snes_common, only: dmda_flow
      use petsc_mpi_common, only : petsc_mpi_finalize, petsc_mpi_barrier
#endif      


      implicit none
#ifdef PETSC
#if (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR >= 6 && PETSC_VERSION_MINOR < 8)
#include <petsc/finclude/petscsys.h>
#elif (PETSC_VERSION_MAJOR >= 3 && PETSC_VERSION_MINOR < 6)
#include <finclude/petscsys.h>
#endif
#endif

#ifdef PETSC
      PetscErrorCode :: ierrcode
#endif

      external checkerr, cvolume, readbloc, xyzcoord

      logical found_section,found_subsection,bflag

      character*72 subsection, strbuffer
      character*256 :: str_usg_generate_file
      
      integer :: i, j, ierr, ixx, iyy, izz, ivtk, l_string, ierrcd
      integer :: nvx_inter, nvy_inter, nvz_inter
      real*8 :: xtot, ytot ! CBF

      ierrcd = 0

!c  choose spatial discretization type, the default is strucutre grid
      discretization_type = 0
      usg_mesh_ordering = 0

!c  flag to indicate the hdf5 global to natural ordering is not set
      b_init_hdf5_lg2g = .false.

!c  radial coordinates option
      radial_coord = .false.  

!c  read spatial discretization parameters and write to temporary file   
   
      section_header = 'spatial discretization'
      call readbloc (idat,itmp,section_header,found_section,.true.)

!c  define length of section header

      l_string = index(section_header,' ')-1
      if (l_string.eq.-1.or.l_string.gt.72) then
         l_string=72
      end if

!c  terminate program if section header not found

      if (.not.found_section) then
        if (rank == 0) then  
          write(*,*) 'error reading input file'
          write(*,*) 'section "',section_header(:l_string),'" missing'
          close(ilog)
        end if
#ifdef PETSC
        call petsc_mpi_finalize
#endif
        stop
      end if

100   continue

      if (discretization_type == 0) then

!c  read spatial discretization parameters and allocate memory

!c  x-direction

        nvxgbl = 0
        ierrcd = 16
        !c to be checked later
        write(*,*) 'start of read nxx'

        read(itmp,*,err=999,end=999) nxx

        !c to be checked later
        write(*,*) 'end of read nxx'

        allocate (xmax(nxx), stat = ierr)
        xmax=0.0d0
        call checkerr(ierr,'xmax',ilog)
        call memory_monitor(sizeof(xmax),'xmax',.true.)

        allocate (xmin(nxx), stat = ierr)
        xmin=0.0d0
        call checkerr(ierr,'xmin',ilog)
        call memory_monitor(sizeof(xmin),'xmin',.true.)

        allocate (nvix(nxx), stat = ierr)
        nvix=0
        call checkerr(ierr,'nvix',ilog)
        call memory_monitor(sizeof(nvix),'nvix',.true.)

        do ixx = 1,nxx
          ierrcd = 17
          read(itmp,*,err=999,end=999) nvix(ixx)
          read(itmp,*,err=999,end=999) xmin(ixx), xmax(ixx)
          nvxgbl = nvxgbl+nvix(ixx)
          xtot=xmax(ixx)-xmin(ixx) ! CBF
        end do

        allocate (delx(nvxgbl), stat = ierr)
        delx=0.0d0
        call checkerr(ierr,'delx',ilog)
        call memory_monitor(sizeof(delx),'delx',.true.)

        allocate (xglat(nvxgbl), stat = ierr)
        xglat=0.0d0
        call checkerr(ierr,'xglat',ilog)
        call memory_monitor(sizeof(xglat),'xglat',.true.)

!c  y-direction

        nvygbl = 0
        read(itmp,*) nyy

        allocate (ymax(nyy), stat = ierr)
        ymax=0.0d0
        call checkerr(ierr,'ymax',ilog)
        call memory_monitor(sizeof(ymax),'ymax',.true.)

        allocate (ymin(nyy), stat = ierr)
        ymin=0.0d0
        call checkerr(ierr,'ymin',ilog)
        call memory_monitor(sizeof(ymin),'ymin',.true.)

        allocate (nviy(nyy), stat = ierr)
        nviy=0
        call checkerr(ierr,'nviy',ilog)
        call memory_monitor(sizeof(nviy),'nviy',.true.)

        do iyy = 1,nyy
          ierrcd = 18
          read(itmp,*,err=999,end=999) nviy(iyy)
          read(itmp,*,err=999,end=999) ymin(iyy), ymax(iyy)
          nvygbl = nvygbl+nviy(iyy)
          ytot=ymax(iyy)-ymin(iyy) ! CBF
        end do

        allocate (dely(nvygbl), stat = ierr)
        dely=0.0d0
        call checkerr(ierr,'dely',ilog)
        call memory_monitor(sizeof(dely),'dely',.true.)

        allocate (yglat(nvygbl), stat = ierr)
        yglat=0.0d0
        call checkerr(ierr,'yglat',ilog)
        call memory_monitor(sizeof(yglat),'yglat',.true.)

!c  z-direction

        nvzgbl = 0
        ierrcd = 19
        read(itmp,*,err=999,end=999) nzz

        allocate (zmax(nzz), stat = ierr)
        zmax=0.0d0
        call checkerr(ierr,'zmax',ilog)
        call memory_monitor(sizeof(zmax),'zmax',.true.)

        allocate (zmin(nzz), stat = ierr)
        zmin=0.0d0
        call checkerr(ierr,'zmin',ilog)
        call memory_monitor(sizeof(zmin),'zmin',.true.)

        allocate (nviz(nzz), stat = ierr)
        nviz=0
        call checkerr(ierr,'nviz',ilog)
        call memory_monitor(sizeof(nviz),'nviz',.true.)

        do izz = 1,nzz
          ierrcd = 20
          read(itmp,*,err=999,end=999) nviz(izz)
          read(itmp,*,err=999,end=999) zmin(izz), zmax(izz)
          nvzgbl = nvzgbl+nviz(izz)
        end do

        allocate (delz(nvzgbl), stat = ierr)
        delz=0.0d0
        call checkerr(ierr,'delz',ilog)
        call memory_monitor(sizeof(delz),'delz',.true.)

        allocate (zglat(nvzgbl), stat = ierr)
        zglat=0.0d0
        call checkerr(ierr,'zglat',ilog)
        call memory_monitor(sizeof(zglat),'zglat',.true.)

!c  calculate total number of control volumes
        nngbl = nvxgbl*nvygbl*nvzgbl

!c  calculate profil area exposed to atmosphere - CBF 

        toparea = xtot*ytot
    

        if (nvxgbl > 1 .and. nvygbl > 1 .and. nvzgbl > 1) then
          cell_projection = projection_xyz
          ndim_sys = 3
        else if (nvxgbl > 1 .and. nvygbl > 1) then
          cell_projection = projection_xy
          ndim_sys = 2
        else if (nvygbl > 1 .and. nvzgbl > 1) then
          cell_projection = projection_yz
          ndim_sys = 2
        else if (nvxgbl > 1 .and. nvzgbl > 1) then
          cell_projection = projection_xz
          ndim_sys = 2
        else if (nvxgbl > 1) then
          cell_projection = projection_x
          ndim_sys = 1
        else if (nvygbl > 1) then
          cell_projection = projection_y
          ndim_sys = 1
        else if (nvzgbl > 1) then
          cell_projection = projection_z 
          ndim_sys = 1 
        end if


#ifdef PETSC
        if (nngbl < nprcs*2) then
          if (rank == 0) then
            write(*,'(a)')                                          &
                'Error: Required minimum subdomain discretization is 2.'
            write(*,'(a)')                                          &
                '       Increase discretization or use less processors.'
            write(*,'(a)')                                             &
                'Error: Required minimum subdomain discretization is 2.'
            write(*,'(a)')                                             &
                '       Increase discretization or use less processors.'
          end if
          call petsc_mpi_finalize
          stop
        end if
#endif
      
!c  domain decomposition
#ifdef PETSC 
        if (varsat_flow) then
          call solver_dd_DMDACreate_flow_heat(.true.)
        end if

        nvx = dmda_flow%xl
        nvy = dmda_flow%yl
        nvz = dmda_flow%zl
        nvxgl = dmda_flow%gxl
        nvygl = dmda_flow%gyl
        nvzgl = dmda_flow%gzl
        nvxls = dmda_flow%xs
        nvxle = dmda_flow%xe
        nvyls = dmda_flow%ys
        nvyle = dmda_flow%ye
        nvzls = dmda_flow%zs
        nvzle = dmda_flow%ze
        nvxgls = dmda_flow%gxs
        nvxgle = dmda_flow%gxe
        nvygls = dmda_flow%gys
        nvygle = dmda_flow%gye
        nvzgls = dmda_flow%gzs
        nvzgle = dmda_flow%gze
#else
        nvx = nvxgbl
        nvy = nvygbl
        nvz = nvzgbl
        nvxgl = nvxgbl
        nvygl = nvygbl
        nvzgl = nvzgbl
        nvxls = 1
        nvxle = nvxgbl
        nvyls = 1
        nvyle = nvygbl
        nvzls = 1
        nvzle = nvzgbl
        nvxgls = 1
        nvxgle = nvxgbl
        nvygls = 1
        nvygle = nvygbl
        nvzgls = 1
        nvzgle = nvzgbl
#endif

        nn = nvx*nvy*nvz
        nngl = nvxgl*nvygl*nvzgl

        if(nvx > 1) then
          nvx_inter = nvx-1+nvxls-nvxgls
        else
          nvx_inter = 1
        end if

        if(nvy > 1) then
          nvy_inter = nvy-1+nvyls-nvygls
        else
          nvy_inter = 1
        end if

        if(nvz > 1) then
          nvz_inter = nvz-1+nvzls-nvzgls
        else
          nvz_inter = 1
        end if

        nn_interfacial = nvx_inter*nvy_inter*nvz_inter
      
      
#ifdef PETSC
        allocate(nn_ranks(nprcs), stat = ierr)
        call checkerr(ierr,'nn_ranks',ilog)
        nn_ranks = 0
        call memory_monitor(sizeof(nn_ranks),'nn_ranks',.true.)

        allocate(nngl_ranks(nprcs), stat = ierr)
        call checkerr(ierr,'nngl_ranks',ilog)
        nngl_ranks = 0
        call memory_monitor(sizeof(nngl_ranks),'nngl_ranks',.true.)

        allocate(nn_interfacial_ranks(nprcs), stat = ierr)
        call checkerr(ierr,'nn_interfacial_ranks',ilog)
        nn_interfacial_ranks = 0
        call memory_monitor(sizeof(nn_interfacial_ranks),'nn_interfacial_ranks',.true.)

        call MPI_Allgather(nn, 1, MPI_INTEGER4, nn_ranks, 1,           &
                           MPI_INTEGER4, Petsc_Comm_World, ierrcode)
        CHKERRQ(ierrcode)

        call MPI_Allgather(nngl, 1, MPI_INTEGER4, nngl_ranks, 1,       &
                           MPI_INTEGER4, Petsc_Comm_World, ierrcode)
        CHKERRQ(ierrcode)

        call MPI_Allgather(nn_interfacial, 1, MPI_INTEGER4,            &
                           nn_interfacial_ranks, 1, MPI_INTEGER4,      &
                           Petsc_Comm_World, ierrcode)
        CHKERRQ(ierrcode)

        nn_offset = 0
        nngl_offset = 0
        nn_interfacial_offset = 0
        do i = 1, rank
          nn_offset = nn_offset + nn_ranks(i)
          nngl_offset = nngl_offset + nngl_ranks(i)
          nn_interfacial_offset = nn_interfacial_offset +              &
                                  nn_interfacial_ranks(i)
        end do
#endif


!c  allocate remaining arrays for spatial discretization 
        allocate (cvol(nngl), stat = ierr)
        cvol=0.0d0
        call checkerr(ierr,'cvol',ilog)
        call memory_monitor(sizeof(cvol),'cvol',.true.)

    
        allocate (xg(nngl), stat = ierr)
        xg=0.0d0
        call checkerr(ierr,'xg',ilog)
        call memory_monitor(sizeof(xg),'xg',.true.)


        allocate (yg(nngl), stat = ierr)
        yg=0.0d0
        call checkerr(ierr,'yg',ilog)
        call memory_monitor(sizeof(yg),'yg',.true.)


        allocate (zg(nngl), stat = ierr)
        zg=0.0d0
        call checkerr(ierr,'zg',ilog)
        call memory_monitor(sizeof(zg),'zg',.true.)


        allocate (dimcv(3,nngl), stat = ierr)
        dimcv=0.0d0
        call checkerr(ierr,'dimcv',ilog)
        call memory_monitor(sizeof(dimcv),'dimcv',.true.)

!c  calculate spatial increments and coordinates of for centroids of 
!c  control volumes
!c  Parallelized, OpenMP, DSU

        call xyzcoord(discretization_type)

#ifdef PETSC
        xlmin = dmda_flow%xlmin
        xlmax = dmda_flow%xlmax
        ylmin = dmda_flow%ylmin
        ylmax = dmda_flow%ylmax
        zlmin = dmda_flow%zlmin
        zlmax = dmda_flow%zlmax
        xlmingbl = dmda_flow%xlmingbl
        xlmaxgbl = dmda_flow%xlmaxgbl
        ylmingbl = dmda_flow%ylmingbl
        ylmaxgbl = dmda_flow%ylmaxgbl
        zlmingbl = dmda_flow%zlmingbl
        zlmaxgbl = dmda_flow%zlmaxgbl
#else
        xlmin = xg(1)
        xlmax = xg(nn)
        ylmin = yg(1)
        ylmax = yg(nn)
        zlmin = zg(1)
        zlmax = zg(nn)
        xlmingbl = xg(1)
        xlmaxgbl = xg(nn)
        ylmingbl = yg(1)
        ylmaxgbl = yg(nn)
        zlmingbl = zg(1)
        zlmaxgbl = zg(nn)
#endif

        if (xglat(nvxgls)< xglat(nvxgle)) then
          gxlmin = xglat(nvxgls)
          gxlmax = xglat(nvxgle)
        else
          gxlmin = xglat(nvxgle)
          gxlmax = xglat(nvxgls)
        end if

        if (yglat(nvygls)< yglat(nvygle)) then
          gylmin = yglat(nvygls)
          gylmax = yglat(nvygle)
        else
          gylmin = yglat(nvygle)
          gylmax = yglat(nvygls)
        end if

        if (zglat(nvzgls)< zglat(nvzgle)) then
          gzlmin = zglat(nvzgls)
          gzlmax = zglat(nvzgle)
        else
          gzlmin = zglat(nvzgle)
          gzlmax = zglat(nvzgls)
        end if

!c  calculate volumes
!c  Parallelized, OpenMP, DSU

        call cvolume

        !c set node flags, used for node selection
        allocate(node_iflags(nngl), stat = ierr)
        call checkerr(ierr,'node_iflags',ilog)
        node_iflags = -1
        call memory_monitor(sizeof(node_iflags),'node_iflags',.true.)

      else

      end if

      goto 1000

999   continue
      if (rank == 0) then
        write(*,*) 'error reading input file, error code ', ierrcd
        write(*,*) 'section "',section_header(:l_string),'"'
        close(ilog)
      end if
#ifdef PETSC
      call petsc_mpi_finalize
#endif
      stop

1000  continue

      return
      end
